# 
# HotQCDEOS.py                                                               
# 
# J. Goswami 
# 

import numpy as np
import sympy as sy
import warnings
from latqcdtools.math.polynomials import Rational
from latqcdtools.math.optimize import persistentSolve

warnings.filterwarnings("error")
warnings.filterwarnings("ignore", category=DeprecationWarning)

pid = 95 * np.pi ** 2 / 180
x0 = 0.9761
ct = 3.8706

T0 = 154

a = np.array([0, -8.7704, 3.9200, 0, 0.3419])
b = np.array([0, -1.2600, 0.8425, 0, -0.0475])

r = 0.4


class EOS:
    def __init__(self, temp):
        self.temp = temp

    def __repr__(self) -> str:
        return "EOS"

    def pressure(self):
        """
        calculation of pressure from the parametrization of 1407.6387
        """
        T = self.temp / T0
        i, ai, bi, x = sy.symbols('i, ai, bi, x')
        exprP = 0.5 * (1 + sy.tanh(ct * (x - x0))) * (pid + sum(a[i] / x**i for i in range(1, 5))) \
                                                   / (1   + sum(b[i] / x**i for i in range(1, 5)))
        pressure = sy.lambdify((ai, bi, x), exprP, modules=[
            'numpy', {'tanh': np.tanh}])(a, b, T)
        return pressure

    def dpressure(self):
        """
        calculation of derivative of pressure from the parametrization of 1407.6387
        """
        T = self.temp / T0
        i, ai, bi, x = sy.symbols('i, ai, bi, x')
        exprP = 0.5 * (1 + sy.tanh(ct * (x - x0))) * (pid + sum(a[i] / x ** i for i in range(1, 5))) \
                / (1 + sum(b[i] / x ** i for i in range(1, 5)))
        exprDerivative = x * sy.diff(exprP, x, 1)
        dPressure = sy.lambdify((ai, bi, x), exprDerivative, modules=[
            'numpy', {'tanh': np.tanh}])(a, b, T)
        return dPressure

    def ObsEoS(self):
        """
        This routine is to calculate pressure , energy density and entropy density for muB = 0
        @return: pressure , energy density and entropy density at muB = 0
        """
        p0 = self.pressure()
        e0 = 3 * p0 + self.dpressure()
        s0 = p0 + e0
        return p0, e0, s0

    def TDerivatives(self, y, pars):
        x = sy.symbols('x')
        a, b, c, d, e, f, g, h, i = sy.symbols('a b c d e f g h i')

        num = a + b * x + c * x ** 2 + d * x ** 3 + e * x ** 4
        den = 1 + f * x + g * x ** 2 + h * x ** 3 + i * x ** 4

        deriv = sy.lambdify((x, a, b, c, d, e, f, g, h, i), sy.diff(num / den, x))

        a0, a1, a2, a3, a4, b1, b2, b3, b4 = pars
        return deriv(y, a0, a1, a2, a3, a4, b1, b2, b3, b4)

    def createCoeffs(self, y, par):
        return Rational([par[0], par[1], par[2], par[3], par[4]], [1, par[5], par[6], par[7], par[8]])(y)

    def ObsEoSfixedmuB(self, Np, qP, muB = 1.0):
        """
        calculation of equation of state( pressure, energy density and
        entropy density for the functional form given in,
        arXiv : 2212.10016 or 1701.04325 [hep-lat]
        @param Np: list of parameters for number density co-efficients
        @param qP: list of parameters for q's
        @param muB: value of the fixed muB
        @return: pressure , energy density and entropy density at fixed muB
        """
        p0 = self.pressure()
        e0 = 3 * p0 + self.dpressure()
        s0 = p0 + e0

        T = T0 / self.temp

        N1B = self.createCoeffs(T, Np[0])
        N3B = self.createCoeffs(T, Np[1])
        N5B = self.createCoeffs(T, Np[2])

        dN1B = (-T) * self.TDerivatives(T, Np[0])
        dN3B = (-T) * self.TDerivatives(T, Np[1])
        dN5B = (-T) * self.TDerivatives(T, Np[2])

        q1 = self.createCoeffs(T, qP[0])
        q3 = self.createCoeffs(T, qP[1])
        q5 = self.createCoeffs(T, qP[2])

        dq1 = (-T) * self.TDerivatives(T, qP[0])
        dq3 = (-T) * self.TDerivatives(T, qP[1])
        dq5 = (-T) * self.TDerivatives(T, qP[2])

        P2B = (N1B + r * q1 * N1B) / 2
        P4B = (N3B + r * (q1 * N3B + 3 * q3 * N1B)) / 4
        P6B = (N5B + r * (q1 * N5B + 3 * q3 * N3B + 5 * q5 * N1B)) / 6

        dP2B = (dN1B + r * (q1 * dN1B + dq1 * N1B)) / 2
        dP4B = (dN3B + r * (dq1 * N3B + q1 * dN3B + 3 * dq3 * N1B + 3 * q3 * dN1B)) / 4
        dP6B = (dN5B + r * (dq1 * N5B + q1 * dN5B + 3 * dq3 * N3B + 3 * q3 * dN3B + 5 * dq5 * N1B + 5 * q5 * dN1B)) / 6

        E2B = (3 * P2B + dP2B - r * dq1 * N1B)
        E4B = (3 * P4B + dP4B - r * (dq1 * N3B + dq3 * N1B))
        E6B = (3 * P6B + dP6B - r * (dq1 * N5B + dq3 * N3B + dq5 * N1B))

        S2B = (4 * P2B + dP2B - N1B - r * ((q1 + dq1) * N1B))
        S4B = (4 * P4B + dP4B - N3B - r * ((q1 + dq1) * N3B + (q3 + dq3) * N1B))
        S6B = (4 * P6B + dP6B - N5B - r * ((q1 + dq1) * N5B + (q3 + dq3) * N3B + (q5 + dq5) * N1B))

        P = p0 + P2B * muB ** 2 + P4B * muB ** 4 + P6B * muB ** 6
        
        NB = N1B * muB + N3B * muB ** 3 + N5B * muB ** 5

        E = e0 + E2B * muB ** 2 + E4B * muB ** 4 + E6B * muB ** 6

        S = s0 + S2B * muB ** 2 + S4B * muB ** 4 + S6B * muB ** 6

        return P, NB, E, S

    def ObsEoSfixedsnB(self, Np, qP, snB = 50):
        """
        calculation of equation of state( pressure,
        energy density and entropy density for the functional
        form given in, arXiv : 2212.10016 or 1701.04325 [hep-lat]
        @param Np: list of parameters for number density co-efficients
        @param qP: list of parameters for q's
        @param snB: For fixed s/nB
        @return: muB/T , pressure , energy density and entropy density at fixed s/nB
        """
        p0 = self.pressure()
        e0 = 3 * p0 + self.dpressure()
        s0 = p0 + e0

        T = T0 / self.temp

        N1B = self.createCoeffs(T, Np[0])
        N3B = self.createCoeffs(T, Np[1])
        N5B = self.createCoeffs(T, Np[2])

        dN1B = (-T) * self.TDerivatives(T, Np[0])
        dN3B = (-T) * self.TDerivatives(T, Np[1])
        dN5B = (-T) * self.TDerivatives(T, Np[2])

        q1 = self.createCoeffs(T, qP[0])
        q3 = self.createCoeffs(T, qP[1])
        q5 = self.createCoeffs(T, qP[2])

        dq1 = (-T) * self.TDerivatives(T, qP[0])
        dq3 = (-T) * self.TDerivatives(T, qP[1])
        dq5 = (-T) * self.TDerivatives(T, qP[2])

        P2B = (N1B + r * q1 * N1B) / 2
        P4B = (N3B + r * (q1 * N3B + 3 * q3 * N1B)) / 4
        P6B = (N5B + r * (q1 * N5B + 3 * q3 * N3B + 5 * q5 * N1B)) / 6

        dP2B = (dN1B + r * (q1 * dN1B + dq1 * N1B)) / 2
        dP4B = (dN3B + r * (dq1 * N3B + q1 * dN3B + 3 * dq3 * N1B + 3 * q3 * dN1B)) / 4
        dP6B = (dN5B + r * (dq1 * N5B + q1 * dN5B + 3 * dq3 * N3B + 3 * q3 * dN3B + 5 * dq5 * N1B + 5 * q5 * dN1B)) / 6

        E2B = (3 * P2B + dP2B - r * dq1 * N1B)
        E4B = (3 * P4B + dP4B - r * (dq1 * N3B + dq3 * N1B))
        E6B = (3 * P6B + dP6B - r * (dq1 * N5B + dq3 * N3B + dq5 * N1B))

        S2B = (4 * P2B + dP2B - N1B - r * ((q1 + dq1) * N1B))
        S4B = (4 * P4B + dP4B - N3B - r * ((q1 + dq1) * N3B + (q3 + dq3) * N1B))
        S6B = (4 * P6B + dP6B - N5B - r * ((q1 + dq1) * N5B + (q3 + dq3) * N3B + (q5 + dq5) * N1B))

        def Psol(x): return p0 + P2B * x ** 2 + P4B * x ** 4 + P6B * x ** 6

        def Esol(x): return e0 + E2B * x ** 2 + E4B * x ** 4 + E6B * x ** 6

        def Ssol(x): return s0 + S2B * x ** 2 + S4B * x ** 4 + S6B * x ** 6

        def NBsol(x): return N1B * x + N3B * x ** 3 + N5B * x ** 5

        def fun(x): return Ssol(x) - snB * NBsol(x)

        xs = np.ones(N1B.size)

        muBsol = persistentSolve(fun, xs)

        return muBsol, Psol(muBsol), NBsol(muBsol), Esol(muBsol), Ssol(muBsol)



# 
# continuumExtrap.py                                                               
# 
# D. Clarke
# 
# A simple module for performing extrapolations to fit functions.
# 

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.plotting import plt
from latqcdtools.statistics.fitting import Fitter
from latqcdtools.base.speedify import DEFAULTTHREADS
from latqcdtools.base.check import checkType


def _powerSeries(x,coeffs):
    """ 
    The default fit form for a continuum extrapolation is a power series in a^2.

    Args:
        x (array-like): a^2 or 1/Nt^2 
        coeffs (array-like): power series coefficients 

    Returns:
        array-like: power series in x 
    """
    result = 0.
    for i in range(len(coeffs)):
        result += coeffs[i]*x**i
    return result


class Extrapolator(Fitter):


    def __init__(self, x, obs, obs_err, order=1, xtype="a", error_strat='propagation', ansatz=None, nproc=DEFAULTTHREADS,
                 tol=1e-12, max_fev=None):
        """ 
        A framework for doing continuum limit extrapolations.

        Args:
            x (array-like): a data or Nt data. (It will get squared in here.) 
            obs (array-like)
            obs_err (array-like)
            ansatz (func, optional): continuum-limit fit ansatz. Power series in a^2 by default.
            order (int, optional): order of the power series. Defaults to 1.
            xtype (str, optional): choose to fit a data or Nt data. Defaults to "a".
            error_strat (str, optional): calculate errors using error propagation or augmented chi^2. Defaults to 'propagation'.
            nproc (int, optional): number of processors for fitting. Defaults to DEFAULTTHREADS.
            tol (float, optional): tolerance for the minimization. Defaults to 1e-12
            max_fev (int, optional): maximum number of iterations. Defaults to 10000
        """
        checkType('int',order=order)
        checkType(np.ndarray,x=x)
        checkType(np.ndarray,obs=obs)
        checkType(np.ndarray,obs_err=obs_err)

        self._order              = order
        self._triedExtrapolation = False
        self._ansatz             = ansatz

        if xtype == "a":
            x = x**2
        elif xtype == "Nt":
            x = 1/x**2
        else:
            logger.TBRaise('Unknown xtype',xtype)
        if order<1:
            logger.TBRaise('Please input order > 1.')

        if ansatz is None:
            ansatz = _powerSeries
        else:
            if self._order != 1:
                logger.warn('Not using a power series ansatz, but still using custom order.')
            logger.debug('received ansatz',ansatz)

        Fitter.__init__(self, ansatz, x, obs, obs_err, norm_err_chi2=False, error_strat=error_strat, nproc=nproc,
                        tol=tol,max_fev=max_fev)


    def __repr__(self) -> str:
        return "Extrapolator"


    def extrapolate(self,start_coeffs=None,prior=None,priorsigma=None,detailedInfo=False,show_results=False):
        """ 
        Carry out the extrapolation.

        Args:
            start_coeffs (array-like, optional): your guess for starting parameters. Defaults to None.
            prior (array-like, optional): Bayesian priors. Defaults to None.
            priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.
            detailedInfo (bool, optional): Do you want information like AIC? Defaults to False.
            show_results (bool, optional): Print fit results to screen? Defaults to False.
            
        Returns:
            (array-like, array-like, float, float): fit result, its error, chi^2/d.o.f., and logGBF if relevant.
        """
        if start_coeffs is None:
            if self._ansatz is not None:
                logger.TBRaise('You need to provide start_coeffs if you use an ansatz.')
            coeffs = np.ones(self._order+1) 
        else:
            checkType(np.ndarray,start_coeffs=start_coeffs)
            coeffs=start_coeffs
        self._triedExtrapolation = True
        if prior is None:
            self._result, self._result_err, self._chidof, self._stats = self.try_fit(start_params=coeffs, algorithms=["curve_fit"], 
                                                                                     detailedInfo=True, show_results=show_results)
        else:
            self._result, self._result_err, self._chidof, self._stats = self.try_fit(start_params=coeffs, priorval=prior, 
                                                                                     priorsigma=priorsigma, show_results=show_results, 
                                                                                     algorithms=["Nelder-Mead"], detailedInfo=True)
        if detailedInfo:
            return self._result, self._result_err, self._chidof, self._stats
        else:
            return self._result, self._result_err, self._chidof


    def plot(self,**kwargs):
        """ 
        Add extrapolation to plot. Accepts the same kwargs as Fitter.plot_fit. 
        """
        if not self._triedExtrapolation:
            logger.TBRaise("Can't plot an extrapolation without having extrapolated first...")
        domain=(1e-8,np.max(self._xdata))
        self.plot_data(**kwargs)
        kwargs['label']=None
        self.plot_fit(domain,**kwargs)


    def save(self,filename,header):
        if not self._triedExtrapolation:
            logger.TBRaise("Can't save an extrapolation without having extrapolated first...")
        domain=(1e-8,np.max(self._xdata))
        self.save_func(filename,domain,header=header)


def continuumExtrapolate(x,obs,obs_err,order=1,show_results=False,plot_results=False,prior=None, start_coeffs=None,priorsigma=None,
                         error_strat='propagation',xtype="a",nproc=DEFAULTTHREADS,detailedInfo=False):
    """ 
    A convenience wrapper for the Extrapolator. 
    """
    ext = Extrapolator(x, obs, obs_err, xtype=xtype, order=order, error_strat=error_strat, nproc=nproc)
    result = ext.extrapolate(start_coeffs=start_coeffs, prior=prior, priorsigma=priorsigma, detailedInfo=detailedInfo, show_results=show_results)
    if plot_results:
        if xtype=="a":
            xlabel = "$a^2$"
        else:
            xlabel = "$1/N_\\tau^2$"
        ext.plot(xlabel=xlabel)
        plt.show()
    return result



# 
# diracFreespectra.py
# 
# J. Goswami
# 
# Python implementation of the spectrum of the free Lattice QCD dirac operators.
#

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.math.math import id
import itertools

logger.warn('This might be broken... Use with caution.')

class GammaMatrix:
    
    """ 
    The 4x4 gamma matrices used in Euclidean quantum field theory. 
    """
    
    @property
    def __repr__(self) -> str:
        return 'GammaMatrix'

    def g(self, i=1):
        if i == 1:
            gamma: np.ndarray = np.array([[0.0, 0.0, 0.0, -1.0j],
                                       [0.0, 0.0, -1.0j, 0.0],
                                       [0.0, 1.0j, 0.0, 0.0],
                                       [1.0j, 0.0, 0.0, 0.0]])
        elif i == 2:
            gamma: np.ndarray = np.array([[0.0, 0.0, 0.0, -1.0],
                                       [0.0, 0.0, 1.0, 0.0],
                                       [0.0, 1.0, 0.0, 0.0],
                                       [-1.0, 0.0, 0.0, 0.0]])
        elif i == 3:
            gamma: np.ndarray = np.array([[0.0, 0.0, -1.0j, 0.0],
                                       [0.0, 0.0, 0.0, 1.0j],
                                       [1.0j, 0.0, 0.0, 0.0],
                                       [0.0, -1.0j, 0.0, 0.0]])
        elif i == 4:
            gamma: np.ndarray = np.array([[0.0, 0.0, -1.0, 0.0],
                                       [0.0, 0.0, 0.0, -1.0],
                                       [-1.0, 0.0, 0.0, 0.0],
                                       [0.0, -1.0, 0.0, 0.0]])
        return gamma

    def g5(self):
        gamma: np.ndarray = np.array([[1.0, 0.0, 0.0, 0.0],
                                   [0.0, 1.0, 0.0, 0.0],
                                   [0.0, 0.0, -1.0, 0.0],
                                   [0.0, 0.0, 0.0, -1.0]])
        return gamma


class DiracOp(GammaMatrix):

    @property
    def __repr__(self) -> str:
        return "DiracOp"

    def __init__(self, Lx, Ly, Lz, Lt, fermion="Wilson"):
        """ 
        Represent Dirac operator on a Euclidean spacetime lattice.

        Args:
            Lx (int)
            Ly (int)
            Lz (int)
            Lt (int)
            fermion (str, optional): Type of fermion discretization. Defaults to "Wilson".
        """
        self.Lx = Lx
        self.Ly = Ly
        self.Lz = Lz
        self.Lt = Lt
        self.fermion = fermion

    def p(self):
        """  
        Computes the momentum values px, py, pz, pt based on the provided lattice extents. 
        """
        if self.Lt == self.Lx:
            px = 2 * np.pi * np.arange(-self.Lx / 2 + 1, self.Lx / 2 + 1, 1) / self.Lx
            py = 2 * np.pi * np.arange(-self.Ly / 2 + 1, self.Ly / 2 + 1, 1) / self.Ly
            pz = 2 * np.pi * np.arange(-self.Lz / 2 + 1, self.Lz / 2 + 1, 1) / self.Lz
            pt = 2 * np.pi * np.arange(-self.Lt / 2 + 1, self.Lt / 2 + 1, 1) / self.Lt
        else:
            px = 2 * np.pi * np.arange(-self.Lx / 2 + 1, self.Lx / 2 + 1, 1) / self.Lx
            py = 2 * np.pi * np.arange(-self.Ly / 2 + 1, self.Ly / 2 + 1, 1) / self.Ly
            pz = 2 * np.pi * np.arange(-self.Lz / 2 + 1, self.Lz / 2 + 1, 1) / self.Lz
            pt = 2 * np.pi * np.arange(-self.Lt / 2 + 1 + 1 / 2, self.Lt / 2 + 1 + 1 / 2, 1) / self.Lt
        return px, py, pz, pt

    def WilsonOp(self, p, mass):
        """ 
        Compute the Wilson operator

        Args:
            p (array-like): momentum vector
            mass (float): scalar mass 

        Returns:
            array-like: Wilson Dirac operator 
        """
        gamma = np.array([self.g(1), self.g(2), self.g(3), self.g(4)])
        term = sum(1j * np.sin(p[i]) * gamma[i] +
                   (1 - np.cos(p[i])) * id(4) for i in range(len(gamma)))
        massterm = mass * id(4)
        return term + massterm

    def DWMobius4D(self, p, mass, M=1, b=1.5, c=0.5, Ls=12):
        """
        Args:
            p (array-like): momentum vector
            mass (float): scalar mass 
            M (float, optional): domain wall parameter. Defaults to 1.
            b (float, optional): domain wall parameter. Defaults to 1.5.
            c (float, optional): domain wall parameter. Defaults to 0.5.
            Ls (int, optional): domain wall parameter. Defaults to 12.

        Returns:
            array-like: Domain wall Dirac operator 
        """
        gamma_5 = self.g5()
        hkNum = (b + c) * self.WilsonOp(p, -M)
        hkDen = 2 * id(4) + (b - c) * self.WilsonOp(p, -M)
        hk = gamma_5 @ hkNum @ np.linalg.inv(hkDen)
        sgnHk = np.matmul((np.linalg.matrix_power(id(4) + hk, Ls) - np.linalg.matrix_power(id(4) - hk, Ls)),
                          np.linalg.inv(np.linalg.matrix_power(id(4) + hk, Ls) + np.linalg.matrix_power(id(4) - hk, Ls)))
        Ddw = (1 + mass) / 2 * id(4) + (1 - mass) / 2 * np.matmul(gamma_5, sgnHk)
        return Ddw

    def eigvalues(self, mass, M=1, b=1.5, c=0.5, Ls=12):
        """
        Args:
            mass (float): scalar mass 
            M (float, optional): domain wall parameter. Defaults to 1.
            b (float, optional): domain wall parameter. Defaults to 1.5.
            c (float, optional): domain wall parameter. Defaults to 0.5.
            Ls (int, optional): domain wall parameter. Defaults to 12.

        Returns:
            array-like: operator eigenvalues 
        """
        eig = []
        px, py, pz, pt = self.p()
        if self.fermion == "Wilson":
            for (pi, pj, pk, pl) in itertools.product(px, py, pz, pt):
                D = self.WilsonOp([pi, pj, pk, pl], mass)
                eig.append(np.linalg.eig(D)[0])
        elif self.fermion == 'DwMobius':
            for (pi, pj, pk, pl) in itertools.product(px, py, pz, pt):
                D = self.DWMobius4D([pi, pj, pk, pl], mass, M, b, c, Ls)
                eig.append(np.linalg.eig(D)[0])
        else:
            logger.warn('Fermion Not implemented yet, here is spectrum for Wilson fermion : option : Wilson , '
                            'DwMobius')
            for (pi, pj, pk, pl) in itertools.product(px, py, pz, pt):
                D = self.WilsonOp([pi, pj, pk, pl], mass)
                eig.append(np.linalg.eig(D)[0])

        return np.array(eig)



# 
# legacy.py                                                               
# 
# D. Clarke, H. Sandmeyer 
# 
# Here we collect "legacy code". This is code that has to meet some requirements:
#   1. It is hard for David to read,
#   2. and therefore hard for him to maintain,
#   3. while not really being used. 
# 
# This module exists for users who still want to utilize some of this legacy
# functionality. Before making something legacy code, please put some kind of
# warning or error somewhere.
# 


import numpy as np
import math
import latqcdtools.base.logger as logger
from latqcdtools.base.utilities import isHigherDimensional
from latqcdtools.base.speedify import parallel_function_eval, DEFAULTTHREADS
from latqcdtools.statistics.statistics import meanArgWrapper
from latqcdtools.base.initialize import DEFAULTSEED


# Suppose the wrapped function is `sum` and `data` is `((1, 2, 3), (4, 5, 6), (7, 8, 9))`. 
# Then reduce_tuple will pass `(1, 4, 7)`, `(2, 5, 8)`, `(3, 6, 9)` to the `sum` function 
# respectively and return `(sum(1, 4, 7), sum(2, 5, 8), sum(3, 6, 9))`.
def reduce_tuple(func):
    def func_wrapper(data, *args, **kwargs):
        if type(data[0]) is tuple:
            retvalue = ()
            for i in range(len(data[0])):
                obj_array = []
                for k in range(len(data)):
                    obj_array.append(data[k][i])
                retvalue += (func(obj_array, *args, **kwargs),)
            return retvalue
        else:
            return func(data, *args, **kwargs)
    return func_wrapper


@reduce_tuple
def std_median(data, axis = 0):
    return np.median(data, axis)


@reduce_tuple
def std_mean(data, axis = 0):
    return np.mean(data, axis)


@reduce_tuple
def std_var(data, axis = 0):
    data = np.asarray(data)
    return np.var(data, axis = axis, ddof = 1)


@reduce_tuple
def std_dev(data, axis = 0):
    data = np.asarray(data)
    return np.std(data, axis = axis, ddof = 1)


@reduce_tuple
def std_err(data, axis = 0):
    data = np.asarray(data)
    return std_dev(data, axis) / np.sqrt(data.shape[axis])


def _pseudo(mean, mean_i, numb_blocks):
    return numb_blocks*mean - (numb_blocks - 1)*mean_i


def _pseudo_val(mean, mean_i, numb_blocks):
    if type(mean) is tuple:
        retvalue = ()
        for i in range(len(mean)):
            retvalue += (_pseudo(np.array(mean[i]), np.array(mean_i[i]), numb_blocks),)
        return retvalue
    else:
        return _pseudo(np.array(mean), np.array(mean_i), numb_blocks)


def _getBlocksize(ndat,nblocks):
    if ndat < nblocks:
        logger.TBError('More blocks than data! nblocks, ndat =',nblocks,ndat)
    elif ndat==nblocks:
        return 1
    else:
        return math.floor(ndat/nblocks)


@reduce_tuple
def dev_by_dist(data, axis=0, return_both_q=False, percentile=68):
    data = np.asarray(data)
    median = np.nanmedian(data, axis)
    numb_data = data.shape[axis]
    idx_dn = max(int(np.floor((numb_data-1) / 2 - percentile/100 * (numb_data-1) / 2)), 0)
    idx_up = min(int(np.ceil((numb_data-1) / 2 + percentile/100 * (numb_data-1) / 2)), numb_data-1)
    sorted_data = np.sort(data - np.expand_dims(median, axis), axis=axis)
    q_l = np.take(sorted_data, idx_dn, axis)
    q_r = np.take(sorted_data, idx_up, axis)
    if return_both_q:
        return np.abs(q_l), np.abs(q_r)
    else:
        return np.max(np.stack((np.abs(q_l), np.abs(q_r)), axis=0), axis=0)


class nimbleJack:
    """ 
    Class allowing for parallelization of the jackknife function. 
    """

    def __init__(self, func, data, nblocks, confAxis, return_sample, args, nproc):

        self._func = func
        self._data = np.array(data)
        if nblocks==1:
            self._nblocks = len(data) 
        else:
            self._nblocks = nblocks
        if not isHigherDimensional(self._data):
            self._confAxis = 0
        else:
            self._confAxis = confAxis
        self._return_sample = return_sample
        self._args = args
        self._nproc = nproc

        # It could be that different observables have different numbers of measurements/configurations. We only
        # allow for that possibility by setting confAxis==1. If your configurations are ragged and along a
        # different axis, you're going to have to handle that yourself for now.
        if self._confAxis == 1:
            self._lengths = [len(self._data[i]) for i in range(len(self._data))]
            self._blocksizes = [_getBlocksize(length, self._nblocks) for length in self._lengths]
        else:
            self._length = self._data.shape[self._confAxis]
            self._blocksize = _getBlocksize(self._length,self._nblocks)

        if self._confAxis == 1:
            self._numb_observe = len(self._data)
            used_data = [self._data[i][:self._nblocks * self._blocksizes[i]] for i in range(self._numb_observe)]
            used_data = np.array(used_data)
        else:
            if self._length < self._nblocks:
                logger.TBError("More blocks than data. length, nblocks=", self._length, self._nblocks)
            swap_data = np.swapaxes(self._data, axis1=self._confAxis, axis2=0)
            used_data = np.swapaxes(swap_data[:self._blocksize*self._nblocks], 0, self._confAxis)
            if self._nblocks==self._length:
                if not np.array_equal(used_data,self._data):
                    logger.TBError('Something went wrong constucting used_data for remove-1 jackknife.')

        # Get initial estimator of the mean
        self._mean = meanArgWrapper(func, used_data, args)
        self._blockval = parallel_function_eval(self.getJackknifeEstimator, range(self._nblocks), nproc=self._nproc)
        self._mean = std_mean(self._blockval)
        self._error = std_err(self._blockval)

    def __repr__(self) -> str:
            return "jackknife"
        
    def getJackknifeEstimator(self, i):
        """ 
        Gets the ith jackknife estimator from throwing away jackknife block i. 
        """
        block_data = []
        if self._confAxis == 1:
            for k in range(self._numb_observe):
                block_data.append(self._data[k][0:i * self._blocksizes[k]])
                block_data[k] = np.concatenate((block_data[k], self._data[k][(i + 1) * self._blocksizes[k]:self._nblocks * self._blocksizes[k]]))
        else:
            # The idea is that the Markov time measurements are along the _confAxis. We construct the block_data
            # by swapping the conf_data to the 0th axis, so that we can cut out the ith configuration block
            # by just appending together using syntax like "swap_data[j]". Then we swap back at the end to restore
            # the original data structure.
            swap_data = np.swapaxes(self._data,axis1=self._confAxis,axis2=0)
            block_data = []
            for j in range(i*self._blocksize):
                block_data.append(swap_data[j])
            for j in range((i+1)*self._blocksize, self._nblocks*self._blocksize):
                block_data.append(swap_data[j])
            block_data = np.swapaxes(np.array(block_data), axis1=0, axis2=self._confAxis)
        mean_i = meanArgWrapper(self._func, block_data, self._args)
        return _pseudo_val(self._mean, mean_i, self._nblocks)

    def getResults(self):
        if self._return_sample:
            return self._blockval, self._mean, self._error
        else:
            return self._mean, self._error


def jackknife(func, data, numb_blocks=20, conf_axis=1, return_sample=False, args=(), nproc=DEFAULTTHREADS):
    """
    Jackknife routine for arbitrary functions. This routine creates the jackknife like blocked subsets of data and
    passes them to the function in the same format as in the input data. So the idea is to write a function that
    computes an observable from a given data set. This function can be put into this jackkife routine and will get the
    jackknifed blocked data as input. Based on the output of the function, the jackkife mean and error are computed.
    The function may return multiple observables that are either scalars or numpy objects. You can pass a
    multidimensional object as data, but the jackknife function has to know from which axis blocks should be removed.
    This is controlled by conf_axis (default = 0 for one dimensional arrays and default = 1 for higher order arrays).
    Look into the __init__ function of the nimbleJack class to see in detail how this is implemented.

        Parameters
        ----------
        func : callable
            The function that calculates the observable

        data : array_like
            Input data

        numb_blocks : integer
            Number of jackknife blocks

        conf_axis : integer, optional, default = 0 for dim(data) = 1 and default = 1 for dim(data) >= 2
            Axis that should be resampled.

        return_sample : boolean, optional, default = False                                           
            Along with the mean and the error also return the results from the individual samples

        args : array_like or dict, default = ()
            optional arguments to be passed to func. If a dictionary they are passed as **args.

        nproc : integer
            Number of threads to use if you choose to parallelize. nproc=1 turns off parallelization.
    """
    jk = nimbleJack(func, data, numb_blocks, conf_axis, return_sample, args, nproc)
    return jk.getResults()


def _autoSeed(seed):
    """ 
    We use seed=None to flag the seed should be automatically chosen. The problem is that we need
    seed to be an integer when enforcing that different bootstrap samples use different seeds. 
    """
    if seed is None:
        return np.random.randint(0,DEFAULTSEED)
    else:
        return seed


def recurs_append(data, sample_data, axis, conf_axis, sample_size, same_rand_for_obs, i, my_seed):
    """ 
    Recursive function to fill the sample. 
    """

    rng = np.random.default_rng(my_seed+i)
    if axis + 1 == conf_axis:
        numb_observe = len(data)
        if sample_size == 0:
            sample_sizes = [ len(j) for j in data ]
        else:
            sample_sizes = [sample_size]*len(data)

        if not same_rand_for_obs:
            randints = [rng.integers(0, len(data[x]), size=sample_sizes[x]) for x in range(numb_observe)]
        else:
            tmp_rand = rng.integers(0, len(data[0]), size=sample_sizes[0])
            randints = [tmp_rand]*numb_observe
        for x in range(numb_observe):
            sample_data.append(np.array(data[x])[randints[x]])
        return

    else:
        for j in range(len(data)):
            sample_data.append([])
            recurs_append(data[j], sample_data[j], axis + 1, conf_axis, sample_size, same_rand_for_obs, i, my_seed)


class nimbleBoot:

    def __init__(self, func, data, numb_samples, sample_size, same_rand_for_obs, conf_axis, return_sample, seed,
                 err_by_dist, args, nproc):

        self._func=func
        self._data=np.array(data)
        self._numb_samples=numb_samples
        self._sample_size=sample_size
        self._same_rand_for_obs=same_rand_for_obs
        self._conf_axis=conf_axis
        self._return_sample=return_sample
        self._seed=_autoSeed(seed)
        self._err_by_dist=err_by_dist
        self._args=args
        self._nproc = nproc 

        if  self._data.ndim == 1:
            self._conf_axis = 0

        self._sampleval = parallel_function_eval(self.getBootstrapEstimator,range(self._numb_samples),nproc=self._nproc,args=(self._seed,))

        if not self._err_by_dist:
            self._mean = std_mean(self._sampleval)
            self._error = std_dev(self._sampleval)
        else:
            self._mean = std_median(self._sampleval)
            self._error = dev_by_dist(self._sampleval)

    def __repr__(self) -> str:
        return "nimbleBoot"

    def getBootstrapEstimator(self,i,my_seed):
        sample_data = []
        if self._conf_axis == 0: # Case of one dimensional array is special
            rng = np.random.default_rng(my_seed+i)
            if self._sample_size == 0:
                sample_size_tmp = len(self._data)
            else:
                sample_size_tmp = self._sample_size
            randints = rng.integers(0, len(self._data), size=sample_size_tmp)
            sample_data = self._data[randints]

        else:
            axis = 0
            recurs_append(self._data, sample_data, axis, self._conf_axis, self._sample_size, self._same_rand_for_obs, i, my_seed)

        sample_data = np.array(sample_data)

        return meanArgWrapper(self._func, sample_data, self._args)

    def getResults(self):
        if self._return_sample:
            return self._sampleval, self._mean, self._error
        else:
            return self._mean, self._error


def bootstr(func, data, numb_samples, sample_size = 0, same_rand_for_obs = False, conf_axis = 1, return_sample = False,
            seed = None, err_by_dist = False, args=(), nproc=DEFAULTTHREADS):
    """
    Bootstrap for arbitrary functions. This routine resamples the data and passes them to the function in the same
    format as in the input. So the idea is to write a function that computes an observable from a given data set. This
    function can be put into this bootstrap routine and will get bootstrap samples as input. Based on the output of the
    function, the bootstrap mean and error are computed. The function may return multiple observables that are either
    scalars or numpy objects. You can pass a multidimensional object as data, but the bootstrap function has to know
    which axis should be resampled which is controlled by conf_axis (default = 0 for one dimensional arrays and
    default = 1 for higher order arrays.)

        Parameters
        ----------
        func : callable
            The function that calculates the observable

        data : array_like
            Input data

        numb_samples : integer
            Number of bootstrap samples

        sample_size : integer, optional, default = 0
            Size of sample

        same_rand_for_obs : boolean, optional, default = False
            Use the same random numbers for each observable accessed by index conf_axis - 1. Please note:
                - Objects that are accessed by an axis >= conf_axis + 1 do always have the same random numbers.
                - Objects that are accessed by axis conf_axis < conf_axis - 1 never share the same random numbers.

        conf_axis : integer, optional, default = 0 for dim(data) = 1 or default = 1 for dim(data) >= 2
            Axis that should be resampled

        return_sample : boolean, optional, default = False
            Along with the mean and the error also return the results from the individual samples

        seed: integer, optional, default = None
            seed for the random generator. If None, the default seed from numpy is used (probably from time)

        same_rand_for_obs : boolean, optional, default = False
            same random numbers per observable

        err_by_dist : boolean, optional, default = False
            Compute the error from the distribution using the median and the 68% quantile

        args : array_like or dict, default = ()
            optional arguments to be passed to func. If a dictionary the are passed as **args.

        nproc : integer
            Number of threads to use if you choose to parallelize. nproc=1 turns off parallelization.
    """
    bts = nimbleBoot(func, data, numb_samples, sample_size, same_rand_for_obs, conf_axis, return_sample, seed,
                     err_by_dist, args, nproc)
    return bts.getResults()


class nimbleGaussianBoot:

    def __init__(self, func, data, data_std_dev, numb_samples, sample_size, same_rand_for_obs, return_sample, seed,
                 err_by_dist, useCovariance, Covariance, args, nproc, asym_err):

        self._func=func
        self._data=np.array(data)
        self._data_std_dev=np.array(data_std_dev)
        self._numb_samples=numb_samples
        self._sample_size=sample_size
        self._same_rand_for_obs=same_rand_for_obs
        self._return_sample=return_sample
        self._seed=_autoSeed(seed)
        self._err_by_dist=err_by_dist
        self._useCovariance=useCovariance
        self._Covariance=Covariance
        self._args=args
        self._numb_observe = len(data)
        self._nproc = nproc 

        self._sampleval=parallel_function_eval(self.getGaussianBootstrapEstimator,range(self._numb_samples),nproc=self._nproc,args=(self._seed,))

        if not self._err_by_dist:
            self._mean = std_mean(self._sampleval)
            self._error = std_dev(self._sampleval)
        else:
            self._mean = std_median(self._sampleval)
            self._error = dev_by_dist(self._sampleval, return_both_q=asym_err)

    def __repr__(self) -> str:
        return "nimbleGaussianBoot"

    def getGaussianBootstrapEstimator(self,i,my_seed):
        sample_data = []
        if not self._same_rand_for_obs:
            for k in range(self._numb_observe):
                rng=np.random.default_rng(my_seed+i+(k+1)*self._numb_samples)
                if not self._useCovariance:
                    if self._sample_size == 1:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k]))
                    else:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k], self._sample_size))
                else:
                    if self._sample_size == 1:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2)))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k]))
                    else:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2), self._sample_size))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k], self._sample_size))
        else:
            rng=np.random.default_rng(my_seed+i)
            for k in range(self._numb_observe):
                if not self._useCovariance:
                    if self._sample_size == 1:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k]))
                    else:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k], self._sample_size))
                else:
                    if self._sample_size == 1:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2)))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k]))
                    else:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2), self._sample_size))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k], self._sample_size))

        sample_data = np.array(sample_data)

        return meanArgWrapper(self._func,sample_data,self._args) 

    def getResults(self):
        if self._return_sample:
            return self._sampleval, self._mean, self._error
        else:
            return self._mean, self._error


def bootstr_from_gauss(func, data, data_std_dev, numb_samples, sample_size = 1, same_rand_for_obs = False,
                       return_sample = False, seed = None, err_by_dist = True, useCovariance = False,
                       Covariance = None, args = (), nproc = DEFAULTTHREADS, asym_err=False):
    """
    Same as standard bootstrap routine, but the data are generated by gaussian noise around the mean values in data.
    The width of the distribution is controlled by data_std_dev. Note, that the function has to average over samples.
    This means that data_std_dev should always be the standard deviation of a single measurement and not the standard
    deviation of a mean.
    """
    data = np.asarray(data)
    data_std_dev = np.asarray(data_std_dev)

    bts_gauss = nimbleGaussianBoot(func=func, data=data, data_std_dev=data_std_dev, numb_samples=numb_samples, 
                                   sample_size=sample_size, same_rand_for_obs=same_rand_for_obs,return_sample=return_sample, 
                                   seed=seed, err_by_dist=err_by_dist, useCovariance=useCovariance, Covariance=Covariance, 
                                   args=args, nproc=nproc, asym_err=asym_err)
    return bts_gauss.getResults()



# 
# polynomials.py                                                               
# 
# J. Goswami
# 

from latqcdtools.base.check import checkType

class Polynomial:

    """
    A simple class for constructing polynomial-type functions. You can construct
    then call a polynomial of only even powers up to fourth order using, for example,
        p = Polynomial([A0, 0., A2, 0. A4])
        p(x)
    """

    def __init__(self, coeffs=None):
        checkType("array",coeffs=coeffs)
        self.coeffs = coeffs

    @property
    def __repr__(self) -> str:
        return "Polynomial{0}".format((str(self.coeffs)))

    def __call__(self, x):
        res = 0.0
        for index_num in range(len(self.coeffs)):
            res += self.coeffs[index_num] *x**index_num
        return res


class Rational:

    """
    A simple class for constructing rational-type functions. 
    """

    def __init__(self, num_coeffs, den_coeffs):
        checkType("array",num_coeffs=num_coeffs)
        checkType("array",den_coeffs=den_coeffs)
        self.num_coeffs = num_coeffs
        self.den_coeffs = den_coeffs

    @property
    def __repr__(self) -> str:
        return "Rational function" + (str(self.num_coeffs) + str(self.den_coeffs))

    def __call__(self, x):
        res = 0.0
        res_den = 0.0
        for index_den in range(len(self.den_coeffs)):
            res_den += self.den_coeffs[index_den] *x**index_den
        for index_num in range(len(self.num_coeffs)):
            res += self.num_coeffs[index_num] *x**index_num/res_den
        return res



# 
# optimize.py
# 
# H. Sandmeyer, D. Clarke
# 
# Some methods related to minimization and Levenberg fits.
#

import scipy.optimize as opt
from scipy.optimize import newton_krylov, fsolve, root
from scipy.optimize._nonlin import NoConvergence
import latqcdtools.base.logger as logger
from latqcdtools.base.check import DivideByZeroError, InvalidValueError, checkType
from latqcdtools.base.utilities import envector 


# This is the base list of exceptions. If encountered, we treat the solve as unreliable.
opt_exceptions = (NoConvergence, FloatingPointError, ValueError, RuntimeWarning, DivideByZeroError, InvalidValueError)


# These orders are not necessarily optimal.
_root_methods  = ['krylov','hybr','broyden1','broyden2','anderson','diagbroyden']
_solve_methods = ['newton_krylov','fsolve','root']


def solve(LHS,guess,tol=1e-8,maxiter=300,method='newton_krylov'):
    """ 
    Wrapper for various methods to solve LHS==0. This is to simplify the interface.

    Args:
        LHS (func)
        guess: initial guess for solution 
        tol (real, optional): Solve tolerance. Defaults to 1e-8.
        maxiter (int, optional): Maximum iterations. Defaults to 300.
        method (str, optional): Defaults to 'newton_krylov'.
    """
    checkType("real",tol=tol)
    checkType('int',maxiter=maxiter)
    if method=='newton_krylov':
        return newton_krylov(LHS, guess, f_tol=tol, f_rtol=tol, maxiter=maxiter)
    elif method=='fsolve':
        return fsolve(LHS, guess, xtol=tol, maxfev=maxiter)
    elif method=='root':
        for root_method in _root_methods:
            try:  
                logger.debug("Trying root:",root_method)
                if root_method in ['hybr']:
                    res = root(LHS, guess, tol=tol, method=root_method, options={'maxfev':maxiter})
                else:
                    res = root(LHS, guess, tol=tol, method=root_method, options={'maxiter':maxiter})
                if res.success == True:
                    return res.x
                else:
                    raise NoConvergence(root_method+' did not converge') 
            except Exception as e:
                logger.debug("Hit exception:",str(e)+"; Terminating.")
                continue
        raise NoConvergence('No root method converged.')
    else:
        logger.TBRaise('Unrecognized method',method)


def persistentSolve(LHS, guess, tol=1e-8, maxiter=300):
    """ 
    Attempt to solve LHS==0 using, in this order, SciPy's newton_krylov, fsolve, and root.

    Args:
        LHS (func)
        guess: initial guess for solution 
        tol (real, optional): Solve tolerance. Defaults to 1e-8.
        maxiter (int, optional): Maximum iterations. Defaults to 300.
    """
    checkType("real",tol=tol)
    checkType('int',maxiter=maxiter)
    for method in _solve_methods:
        try:
            return solve(LHS,guess,tol=tol,maxiter=maxiter,method=method)
        except opt_exceptions as e:
            logger.debug("Hit exception:",e)
        except Exception as e:
            raise e
    raise NoConvergence('No method converged!') 


def minimize(func, jac=None, hess=None, start_params=None, tol=1e-12, maxiter=10000, algorithm=None):
    """ 
    Wrapper for scipy.optimize.minimize. Helps that all algorithms have common syntax.

    Args:
        func: to-be-minimized function 
        jac (func, optional): Explicit Jacobian. Defaults to None.
        hess (func, optional): Explicit Hessian. Defaults to None.
        start_params (array-like, optional): Starting guess. Defaults to None.
        tol (float, optional): Solve tolerance. Defaults to 1e-12.
        maxiter (int, optional): Maximum number of solve iterations. Defaults to 10000.
        algorithm (str, optional): Solve algorithm. Defaults to Scipy default, usually BFGS.

    Raises:
        e: scipy.optimize exception 
        ValueError: minimizer fails to converge

    Returns:
        array-like: solution vector 
    """

    kwargs = {'method': algorithm, 'tol': tol}

    logger.details('Trying',algorithm,'with maxiter=',maxiter)

    if algorithm in ["SLSQP","L-BFGS-B","CG"]:
        kwargs['jac'] = jac
        kwargs['options'] = {'maxiter': maxiter}

    elif algorithm == "TNC":
        kwargs['jac'] = jac
        kwargs['options'] = {'maxfun': maxiter}

    elif algorithm in ["COBYLA","Nelder-Mead"]:
        kwargs['options'] = {'maxiter': maxiter}

    elif algorithm == "Powell":
        kwargs['options'] = {'xtol': tol, 'ftol': tol, 'maxfev': maxiter}

    else:
        kwargs['jac'] = jac
        kwargs['hess'] = hess
        kwargs['options'] = {'maxiter': maxiter}

    if algorithm != "nonlin":
        try:
            res = opt.minimize(func, start_params, **kwargs)
        except Exception as e:
            raise e
    else:
        try:
            res = opt.least_squares(func, start_params, max_nfev=maxiter, ftol=1e-15, xtol=tol, gtol=tol, 
                                    jac=jac,  method='trf', loss='huber')
        except Exception as e:
            raise e

    logger.debug(res)

    params = res.x
    if not res.success:
        logger.details(algorithm, res.message)
        raise ValueError('Minimizer failed.')

    return envector(params)



#
# fitting.py
#
# H. Sandmeyer, D. Clarke
#
# Fitter class, along with functions for fitting data. The goal is to be able to carry out simple fits, trying many
# fitting algorithms. Bayesian priors are supported. You can judge the quality of your fit by quantities like
# chi^2/d.o.f. and logGBF. 
#

import numpy as np
from scipy.optimize import curve_fit
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkEqualLengths
from latqcdtools.base.speedify import DEFAULTTHREADS, parallel_function_eval
from latqcdtools.base.plotting import plot_dots, plot_bar, plt
from latqcdtools.base.readWrite import writeTable
from latqcdtools.base.utilities import envector, isHigherDimensional, toNumpy, createFilePath
from latqcdtools.math.math import invert, regulate, checkSquare, isSymmetric, isPositiveSemidefinite, \
    isMatrix, isVector
from latqcdtools.math.optimize import minimize
from latqcdtools.math.num_deriv import diff_jac, diff_fit_grad
from latqcdtools.statistics.statistics import plot_func, error_prop_func, cov_to_cor, chisquare, logGBF, DOF, \
    expandArgs, checkDomain, BAIC, AIC, AICc, checkPrior, goodnessOfFit
from latqcdtools.base.printErrorBars import get_err_str


# Allowed keys for the constructor
_allowed_keys = ['grad', 'args', 'grad_args', 'tol', 'use_diff', 'error_strat', 'nowarn',
                'norm_err_chi2', 'derive_chisq', 'svdcut', 'test_tol', 'max_fev', 'nproc']

# All possible algorithms.
all_algs = ["curve_fit", "L-BFGS-B", "TNC", "Powell", "Nelder-Mead", "COBYLA", "SLSQP", "CG","dogleg", "trust-ncg", "nonlin"]

# Standard algorithms for the minimization. 
std_algs = ["curve_fit", "TNC", "Powell", "Nelder-Mead", "nonlin"]

# Fast algorithms that work with priors. 
bayes_algs = ["TNC", "Powell", "Nelder-Mead", "nonlin"]


def zipXYData(xdata,ydata):
    """ Collect 1d xdata and ydata into an 2d xydata array. You can then use
    unzipXYData inside of some func(xydata), which represents some f(x,y), to
    separate the x part and y part.

    Args:
        xdata (array-like)
        ydata (array-like)

    Returns:
        np.ndarray: array of x,y coordinates [ (x1,y1), (x2,y1), ..., (x1,y2), ... ]
    """
    x,y = np.meshgrid(xdata,ydata)
    return np.stack((x,y), axis=2).reshape(-1, 2)


def unzipXYData(xydata):
    """ Take a 2d xydata array, created by zipXYData, and extract xvalues and yvalues
    for use inside of a function of two variables.

    Args:
        xydata (np.ndarray): array of x,y coordinates [ (x1,y1), (x2,y1), ..., (x1,y2), ... ]

    Returns:
        xvalues [x1, x2, ... , xN, x1, x2, ...],
        yavlues [y1, y1, ... , y1, y2, y2, ...]
    """
    X, Y = np.hsplit(xydata, 2)
    return np.ravel(X), np.ravel(Y)


class Fitter:
    """ 
    The :class:`Fitter`, contains all information necessary for fitting: The data, the function to be fitted, and
    optional the data for the errors. There are different minimization algorithms available. Many of them need the
    gradient or hessian of the chisquare. One way is to set the derivatives of the fitting function from outside.
    The derivatives of the actual chisquare are then computed via error propagation. Another way is to use numerical
    derivatives.

    There are two ways to compute the derivatives of the chisqare numerically. Either compute the
    numerical derivative of the whole chisquare (error_strat='hessian') or compute the derivatives of the fitting
    function and use error propagation (error_strat='propagation'). The latter is the default case.
    """

    def __init__(self, func, xdata, ydata, edata = None, **kwargs):
        """
        Parameters
        ----------
        func : callable
            Function to be fitted. Must have form 
                func(x, params, *args)
        xdata : array_like
            xdata used for fit. These data may be higher dimensional. This may be the case when our fit functions needs
            more than one parameter. However, the number of elements in the first axis has to be equal to the number of
            elements in ydata.
        ydata : array_like
            ydata used for fit.
        edata : array_like, optional, default: None
            Data for the error. Either pass an 1D array of errors of a full covariance matrix. In case of errors, the
            errors are interpreted as edata = sqrt(variance). For the case of the covariance matrix no root has to be
            taken: variance = diag(edata).
        grad : callable, optional, default: None
            gradient of the fit function.
        args : array_like, optional, default: ()
            Optional arguments that shall be passed to func and that should not be fitted.
        grad_args : array_like, optional, default: None
            Optional parameter for the gradient. If set to None the arguments for the function are used (args).
        tol : float, optional, default: 1e-12
            Tolerance for the minimization.
        svdcut : float, optional, default: 1e-12
            Apply this cut to the eigenvalues of the correlation matrix.
        max_fev : int, optional, default: 10000
            Maximum number of iterations / function evaluations.
        use_diff : bool, optional, default: True
            In case of numerical derivative use the difference quotient for approximation.
        norm_err_chi2 : bool, optional, default: False 
            Multiply errors with chi**2/d.o.f. Some people like to do this. 
        derive_chisq : bool, optional, default: False
            In case of numerical derivative, apply the derivative to the whole chisquare instead of the function.
        nproc : int, optional, default: DEFAULTTHREADS
            If you want you can accelerate the fits using nprocs threads.
        nowarn: bool, optional, default: False
            If you want to turn off warnings.
        
        """

        diff = set(set(kwargs.keys()) - set(_allowed_keys))
        if len(diff) != 0:
            logger.TBRaise("Illegal argument(s) to fitter", *diff)

        # Some attributes that are set in functions other than __init__.
        self._grad = None
        self.grad  = None
        self._pcov = None

        # Store data
        checkEqualLengths(xdata,ydata)
        self._xdata, self._ydata = toNumpy(xdata, ydata) 

        # These attributes are described in the above doccumentation. If they aren't specified in the keyword
        # arguments when the Fitter is initialized, they take the default value shown here. 
        self._use_diff      = kwargs.get('use_diff', True)
        self._derive_chisq  = kwargs.get('derive_chisq', False)
        self._tol           = kwargs.get('tol', 1e-12)
        self._test_tol      = kwargs.get('test_tol', 1e-10)
        self._svdcut        = kwargs.get('svdcut', 1e-12)
        self._max_fev       = kwargs.get('max_fev', None)
        self._norm_err_chi2 = kwargs.get('norm_err_chi2', False)
        self._args          = kwargs.get('args', ())
        self._grad_args     = kwargs.get('grad_args', None)
        self._errorAlg      = kwargs.get('error_strat', 'propagation')
        self._nproc         = kwargs.get('nproc', 1)
        self._nowarn        = kwargs.get('nowarn', False)
        logger.debug('Initialize Fitter with:')
        logger.debug('       use_diff:',self._use_diff) 
        logger.debug('   derive_chisq:',self._derive_chisq)
        logger.debug('            tol:',self._tol)
        logger.debug('       test_tol:',self._test_tol)
        logger.debug('        max_fev:',self._max_fev)
        logger.debug('  norm_err_chi2:',self._norm_err_chi2)
        logger.debug('           args:',self._args)
        logger.debug('      grad_args:',self._grad_args)
        logger.debug('       errorAlg:',self._errorAlg)
        logger.debug('         svdcut:',self._svdcut)
        logger.debug('          nproc:',self._nproc) 

        if (self._errorAlg=='hessian') and (self._derive_chisq) and (not self._nowarn):
            self._errorAlg='propagation'
            logger.warn('hessian strategy not yet compatible with derive_chisq; switching to propagation.')

        if self._grad_args is None:
            self._grad_args = self._args

        if type(self._max_fev) is int:
            tmp_fev = self._max_fev
            self._max_fev = dict()
            for alg in all_algs:
                self._max_fev[alg] = tmp_fev
        if self._max_fev is None:
            self._max_fev = {
                "curve_fit"  : 50000,
                "L-BFGS-B"   : 15000,
                "TNC"        : 5000 ,
                "Powell"     : 10000,
                "Nelder-Mead": 10000 ,
                "COBYLA"     : 15000,
                "SLSQP"      : 15000,
                "CG"         : 15000,
                "dogleg"     : 15000,
                "trust-ncg"  : 15000,
                "nonlin"     : 10000,
                }

        # Initialize func. This is also done in set_func, but we need it before that
        self._func = func

        # The final parameters and covariance matrices will be saved in this. 
        self._saved_params = None 
        self._saved_pcov   = None 

        self.set_func(func, kwargs.get('grad', None)) 

        # For Bayesian fits. If these are used, they are set in general_fit, and then utilized elsewhere.
        self._priorval   = None 
        self._priorsigma = None 

        # Check if we have the covariance matrix available. 
        if edata is not None:
            edata = np.array(edata)
            if isMatrix(edata):
                self._cov=edata
            elif isVector(edata):
                self._cov=np.diag(edata**2)
            else:
                logger.TBRaise('Expected ndim < 3. Got ndim =',edata.ndim)
        else: 
            # Initialize everything to one if we don't get error information. 
            self._cov = np.diag(np.ones(len(self._ydata)))

        checkSquare(self._cov)
        if not isSymmetric(self._cov):
            logger.TBRaise('Covariance matrix is not symmetric.')
        if not isPositiveSemidefinite(self._cov):
            logger.TBRaise('Covariance matrix is not positive semidefinite.')

        # Compute a correlation matrix. Note that if the incoming data are highly correlated, then
        # the covariance matrix is likely to be ill-conditioned. This results in numerical stability
        # issues, most obviously when trying to invert the matrix, but also when trying to do things
        # like compute its jacobian. Therefore regulate it. 
        self._cov     = regulate( self._cov, svdcut=self._svdcut )
        self._cor     = cov_to_cor( self._cov )
        self._inv_cor = invert( self._cor, 'scipy' )
        self._inv_cov = invert( self._cov, 'scipy' )


    def __repr__(self) -> str:
        return "Fitter"


    def set_func(self, func, grad = None, args = None, grad_args = None, hess_args = None):
        """ 
        Set fitting function, gradient, and their arguments. Also initialize self.func,
        and self.grad. These point to the actual wrappers which are used in the fit. In
        case of provided gradient, this will be wrap_grad. In case of numerical derivatives, 
        this will be num_grad.

        Parameters
        ----------
        func : callable
            Function to be fitted.
        grad : callable, optional, default: None
            gradient of the fit function.
        args : array_like, optional, default: ()

        grad_args : array_like, optional, default: None
            Optional parameter for the gradient. If set to None the arguments for the function
            are used (args).
        """

        # Direct storage of the user functions.
        self._func = func
        self._grad = grad

        # Later we only access self.func and self.grad. These are wrappers around the user function or the
        # numerical derivatives. The code below chooses the right wrappers. When they are set to None, None will
        # ultimately be passed to scipy.optimize.minimize, which indicates that minimize should estimate these
        # quantities with numerical derivatives.
        if self._grad is None:
            if self._use_diff and not self._derive_chisq:
                self.grad = self.num_grad
            else:
                self.grad = None
                self._derive_chisq = True
        else:
            self.grad = self.wrap_grad

        if args is not None:
            self._args = args

        if grad_args is not None:
            self._grad_args = grad_args
        elif self._args is not None:
            self._grad_args = self._args


    def num_grad(self, x, params):
        return diff_fit_grad(x, params, self._func, self._args)


    def wrap_grad(self, x, params):
        return expandArgs(self._grad,x,params,self._grad_args)
    def wrap_func(self, x, params):
        return expandArgs(self._func,x,params,self._args)


    def fit_ansatz_array(self, params):
        """ 
        Return the array of the fit ansatz values at each position in self._xdata. 
        """
        params = np.array(params)
        return self.wrap_func(self._xdata, params)


    def jacobian_fit_ansatz_array(self, params):
        """ 
        If f is the fit function, return the array df / dp_i evaluated at each _fit_xdata. 
        """
        params = np.array(params)
        # Note that the function is passed in secret here.
        return self.grad(self._xdata, params).T


    def calc_chisquare(self, params):
        """ 
        Compute chi^2. This is the function that will be minimized. 
        """
        return chisquare(self._xdata, self._ydata, self._cov, self._func, self._args, params, 
                         prior=self._priorval,priorsigma=self._priorsigma)


    def grad_chisquare(self, params):
        """ 
        Compute the gradient of the chisquare, see e.g. eq. (A3) in 10.1103/PhysRevD.90.054506. 
        We assume priors are not correlated with data. 
        """
        jac  = self.jacobian_fit_ansatz_array(params).T     # df/dp
        diff = self.fit_ansatz_array(params) - self._ydata  # Dy
        res  = 2 * jac @ self._inv_cov @ diff
        if self._priorsigma is not None:
            res += np.sum( 2*( np.array(params) - self._priorval ) / self._priorsigma**2 )
        return res


    def hess_chisquare(self, params):
        """ 
        Compute the Hessian of the chisquare. Terms of O(f(p)-y) are discarded, which is reasonable when
        the fit is relatively good. This also ensures non-negative eigenvalues for pcov and saves computational
        time computing the Hessian. See e.g. eq. (A4) in 10.1103/PhysRevD.90.054506. We assume priors are not 
        correlated with data. 
        """
        jac = self.jacobian_fit_ansatz_array(params).T
        res = 2*( jac @ self._inv_cov @ jac.T )
        if self._priorsigma is not None:
            res += np.sum( 2/self._priorsigma**2 )
        return res


    def _num_func_jacobian(self, params):
        """ 
        For the error computation we need the Jacobian of the array of function values. If self._derive_chisq is True,
        we cannot use self.grad_fit_ansatz_array. In that case, the Jacobian is calculated using this function. 
        """
        return diff_jac(params, self.fit_ansatz_array)


    def minimize_chi2(self, start_params, algorithm):
        """ 
        Minimize the chi^2 using the scipy minimize routine.

        Parameters
        ----------
        start_params: array_like
            The start parameters that are used for the minimization.
        algorithm: string
            The algorithm that will be used.

        Returns
        -------
        params:
            Array of the parameters that minimize the chisquare.
        chi2:
            The minimum value of the chisquare.
        """

        if self.grad is not None:
            jac_func=self.grad_chisquare
            hess_func=self.hess_chisquare
        else:
            jac_func=None
            hess_func=None

        if algorithm == "curve_fit":

            if self._priorsigma is not None:
                logger.TBRaise('The curve_fit algorithm is not yet able to handle priors.')

            # This ensures that however func is used, whatever is passed to him as arguments will be captured as a
            # tuple, which is then plugged into wrap_func.
            def func(x, *p):
                return self.wrap_func(x, p)

            cov = self._cov
            # If no gradient has been provided by the user, it is probably better to use the numerical derivative from
            # curve_fit instead of our own.
            if self._grad is not None:
                def grad(x, *p):
                    return np.array(self.grad(x, p)).T
            else:
                grad = None

            params, _ = curve_fit(func, self._xdata, self._ydata, sigma = cov, p0 = start_params, jac = grad,
                                  ftol = self._tol, maxfev = self._max_fev["curve_fit"], method='trf')

        else:
            params = minimize(self.calc_chisquare, jac_func, hess_func, start_params, self._tol,
                              self._max_fev[algorithm], algorithm = algorithm)

        if any(np.isnan(params)) or any(np.isinf(params)):
            logger.TBFail(algorithm,": Fit result is inf or nan!")
            raise ValueError

        chi2 = self.calc_chisquare(params)

        return params, chi2


    def compute_err(self, params, chi2, algorithm):
        """ 
        Compute the covariance matrix of the fit parameters. If no errors have been provided, they are assumed to
        be one. We get the fit variances from the diagonal elements of the covariance matrix.

        Parameters
        ----------
        params: array_like
            Parameters for which the errors should be computed.
        chi2: float
            The chisquare for those parameters.
        algorithm: string
            The algorithm used to compute these errors (Necessary for strings in exceptions).

        Returns
        ------
        pcov:
            Correlation matrix of the fit parameters.
        fit_errors:
            Diagonal of pcov.
        """

        dof = DOF(len(self._ydata),len(params),self._priorsigma)    

        logger.debug('using error alg',self._errorAlg)
        if dof <= 0:
            pcov = np.full((len(params),len(params)), np.nan)
        else:
            if self._errorAlg=='propagation':
                pcov = self.pcov_error_prop(params,algorithm)
            elif self._errorAlg=='hessian':
                pcov = self.pcov_hessian(params,algorithm)
            else:
                logger.TBRaise('Unknown fitting algorithm', self._errorAlg)

        # Sometimes people like to rescale the parameter covariance matrix by the chi^2/dof. This tries to
        # take the fit quality in account into the error directly, and to my understanding it is what
        # gnuplot does. In a physics context, this procedure seems to be unjustified, and it doesn't come 
        # naturally out of the mathematics, so this is not done by default.
        if self._norm_err_chi2:
            pcov *= chi2/dof
        fit_errors = np.sqrt(np.diag(pcov))

        self._pcov = pcov

        return pcov, fit_errors


    def pcov_error_prop(self, params, algorithm):
        """ 
        Compute the parameter's covariance matrix through error propagation, i.e. pcov = (J^t * C^-1 * J)^-1, where
        J is the Jacobian of the fit function and C is the covariance matrix of the data points. 
        """
        if self.grad is None:
            jac = self._num_func_jacobian(params)
        else:
            jac = self.jacobian_fit_ansatz_array(params)

        jej = jac.T @ self._inv_cov @ jac

        try:
            pcov = invert(jej,'pinv')

            # Test that the inversion went relatively OK
            test = pcov @ jej
            if abs(np.sum(test) - np.sum(np.diag(test))) > self._test_tol:
                if not self._nowarn:
                    logger.warn(algorithm,"pcov @ jej not close to id")
                    logger.warn(np.diag(test))

            if np.min(np.diag(pcov)) < 0:
                logger.TBFail(algorithm + ": Negative entries for the variance!")
                raise ValueError

        except Exception as e:
            logger.TBFail(algorithm,"hit exception",e)
            raise e 

        return pcov 


    def pcov_hessian(self, params, algorithm):
        """ 
        Obtain the parameter's covariance matrix by inverting the hessian of the chi^2. 
        """
        pcov = invert(self.hess_chisquare(params)/2.)
        if np.min(np.diag(pcov)) < 0:
            logger.TBFail(algorithm + ": Negative entries for the variance!")
            raise ValueError
        return pcov


    def _general_fit(self, algorithm="curve_fit"): 
        """ 
        Perform fit. No new fit data are generated. 
        """
        params, chi2 = self.minimize_chi2(self._saved_params, algorithm)
        pcov, fit_errors = self.compute_err(params, chi2, algorithm)
        return params, fit_errors, chi2, pcov


    def _tryAlgorithm(self,algorithm):
        """ 
        Wrapper that collects general fit results. Allows for parallelization. 
        """
        try:
            params, fit_errors, chi2, pcov = self._general_fit(algorithm) 
            logger.details(algorithm, "successful. Chi^2 = ", chi2)
            return params, fit_errors, chi2, pcov, None
        except Exception as e:
            logger.details(algorithm, "failed with:", e)
            return None, None, np.inf, None, e


    def _autoDomain(self,domain):
        """ 
        Set domain automatically if domain=None. 
        """
        if domain is None:
            domain=(np.min(self._xdata),np.max(self._xdata))
        checkDomain(domain)
        return domain


    def try_fit(self, algorithms = std_algs, start_params = None, priorval = None, priorsigma = None, detailedInfo = False, show_results = False):
        """ 
        Perform the fit. This is what you should usually call. Try different algorithms and choose the one with the
        smallest chi^2. By default this method does a standard statistical fit. One can also include priors to obtain
        posteriors using Bayesian statistics. A well known summary of the latter strategy in the context of lattice QCD
        is given here: https://arxiv.org/abs/hep-lat/0110175. If there are priors, calculate logGBF and return that too.

        Parameters
        ----------
        algorithms: string or None, optional, default = None
            List of strings with the algorithms that can be used. Possible values are:
                "L-BFGS-B", "TNC", "Powell" ,"Nelder-Mead", "COBYLA", "SLSQP", "CG", "BFGS", "dogleg", "trust-ncg".
            The latter 4 usually don't work. When provided None, the first 7 algorithms are used.
        start_params: array_like, optional, default = None
            The start parameters for the fit.
        priorval: array_like, optional, default = None
            For constrained fits. Prior mean values for the fit.
        priorsigma: array_like, optional, default = None
            For constrained fits. Prior error bars for the fit.
        detailedInfo : bool, optional. default = False
            If True, return also the covariance matrix of the fit parameters and logGBF
        show_results : bool, optional. default = False
            If True, print fit results to screen nicely. 

        Returns
        -------
        params:
            The final fit parameters.
        params_err:
            The error of this fit parameters.
        chidof:
            chi^2/dof.
        stats:
            If detailedInfo, return dictionary with logGBF, BAIC, pcov.
        """

        if isinstance(algorithms,str):
            algorithms=[algorithms]

        if ('curve_fit' in algorithms) and (self._errorAlg=='hessian') and (not self._nowarn):
            self._errorAlg='propagation'
            logger.warn('hessian strategy not yet compatible with curve_fit; switching strategy to propagation.')

        # Prior values are a good point for starting the fit, if you don't have another guess.
        if (priorval is not None) and (start_params is None):
            start_params = priorval

        # For a fit with only one parameter, we also accept a scalar. Check if this is the case.
        if start_params is not None:
            self._saved_params = envector(start_params)

        # Initialize prior values.
        checkPrior(priorval,priorsigma)
        if priorsigma is not None:
            self._priorsigma = np.array(priorsigma)
            self._priorval = np.array(priorval)

        # See if a standard function evaluation works
        try:
            _test = self.wrap_func(self._xdata, self._saved_params)
        except Exception as e:
            logger.TBRaise('Fit function must have signature func(xdata,params,args).',exception=e)
        checkEqualLengths(_test, self._xdata)

        resultSummary  = parallel_function_eval( self._tryAlgorithm, algorithms, nproc=self._nproc )
        all_params     = [row[0] for row in resultSummary]
        all_fit_errors = [row[1] for row in resultSummary]
        all_chi2       = [row[2] for row in resultSummary]
        all_pcov       = [row[3] for row in resultSummary]
        all_except     = [row[4] for row in resultSummary]

        # Check to see if all the fits failed
        if np.all(np.array(all_chi2) == np.inf): 
            for i, algorithm in enumerate(algorithms):
                logger.TBFail(algorithm+": ",all_except[i])
            logger.TBRaise("No algorithm converged. See above list of exceptions.")

        # Find the smallest chi^2
        min_ind = np.argmin(all_chi2)
        if len(algorithms) > 1:
            logger.details("Choosing", algorithms[min_ind], "with chi^2 =", all_chi2[min_ind])

        # Store results internally
        self._saved_params = all_params[min_ind]
        self._saved_pcov = all_pcov[min_ind]

        for i in range(len(algorithms)):
            logger.debug(algorithms[i]+':',all_params[i],all_fit_errors[i],all_chi2[i],all_pcov[i],all_except[i])
            
        dof = DOF(len(self._ydata),len(self._saved_params),self._priorsigma)    

        if dof <= 0:
            chidof = np.inf
        else:
            chidof = all_chi2[min_ind]/dof

        if show_results:
            logger.info('Fit Results:')
            for i in range(len(all_fit_errors[min_ind])):
                logger.info(f'          p[{i}] =',get_err_str(self._saved_params[i],all_fit_errors[min_ind][i]))
            logger.info('  chi^2/d.o.f. =',chidof)


        if detailedInfo:
            return ( np.copy(self._saved_params),
                     all_fit_errors[min_ind],
                     chidof,
                     {  'logGBF'   : logGBF(self._xdata, self._ydata, self._cov, self._func, self._args, self._saved_params,
                                           prior=self._priorval,priorsigma=self._priorsigma),
                        'pcov'     : np.copy(self._saved_pcov),
                        'chi2'     : all_chi2[min_ind],
                        'chi2data' : chisquare(self._xdata, self._ydata, self._cov, self._func, self._args, self._saved_params,
                                               None, None),
                        'BAIC'     : BAIC(self._xdata, self._ydata, self._cov, self._func, self._args, self._saved_params),
                        'AIC'      : AIC(self._xdata, self._ydata, self._cov, self._func, self._args, self._saved_params,
                                         self._priorval, self._priorsigma),
                        'AICc'     : AICc(self._xdata, self._ydata, self._cov, self._func, self._args, self._saved_params,
                                          self._priorval, self._priorsigma),
                        'DOF'      : dof,
                        'Q'        : goodnessOfFit(dof,all_chi2[min_ind])
                     }
                   )
        else:
            return ( np.copy(self._saved_params),
                     all_fit_errors[min_ind],
                     chidof
                   )


    def do_fit(self, algorithm="curve_fit", **kwargs):
        """ 
        Same as try_fit but with only one algorithm. 
        """
        return self.try_fit([algorithm], **kwargs)


    def save_func(self, filename, domain = None, no_error=False, header=None, npoints=1000, **kwargs):
        """ 
        Save fit data to table. 
        """
        domain = self._autoDomain(domain)
        params_err = self._saved_pcov
        params = self._saved_params
        def func(x, params):
            return self._func(x, params, *self._args)
        if no_error:
            save_func(func, filename, domain=domain, args=params, header=header, npoints=npoints,**kwargs)
        else:
            def grad(x, params):
                return np.asarray(self.grad(x, params))
            save_func(func, filename, domain=domain, args=params, args_err=params_err, grad=grad,
                      header=header, npoints=npoints, **kwargs)


    def plot_fit(self, domain = None, no_error = False, **kwargs):
        """ 
        Plot the fit function. 
        """
        if isHigherDimensional(self._xdata):
            logger.TBRaise('Automatic fit plotting only supported for functions of a single variable.') 
        domain = self._autoDomain(domain)
        if not no_error:
            plot_func(self._func, domain=domain, params=self._saved_params, params_err=self._saved_pcov, 
                      grad=self._grad, args=self._args, **kwargs)
        else:
            plot_func(self._func, domain=domain, params=self._saved_params, args=self._args, **kwargs)


    def plot_data(self, **kwargs):
        """ 
        Plot the fit data. 
        """
        if isHigherDimensional(self._xdata):
            logger.TBRaise('Automatic data plotting only supported for functions of a single variable.') 
        if self._cov is not None:
            sigma = np.sqrt( np.diag(self._cov) )
            plot_dots(self._xdata, self._ydata, sigma, **kwargs)
        else:
            plot_dots(self._xdata, self._ydata, **kwargs)


    def plot_cor(self, title = 'Data correlation matrix', xlabel='$x_j$', ylabel='$x_i$'):
        """ 
        Plot the correlation matrix of the input data. 
        """

        ncov = self._fit_cor

        if title is not None:
            plt.title(title)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        xrange = np.arange(len(ncov) + 1)
        yrange = np.arange(len(ncov) + 1)

        plt.pcolormesh(xrange, yrange, ncov, cmap="Blues")
        plt.gca().invert_yaxis()
        cb = plt.colorbar()
        cb.ax.tick_params(labelsize=12)


    def plot_eig(self, xlabel="$i$", ylabel="$\\lambda_i$", title = 'Eigenvalues of data correlation matrix'):

        vals, _ = np.linalg.eig(self._fit_cor)

        eig_real = np.real(vals)
        eig_imag = np.imag(vals)

        plot_bar(range(len(eig_real)), eig_real, color='#d32d11', label="real", alpha=0.7, title=title, xlabel=xlabel, ylabel=ylabel)
        if np.min(eig_imag) != 0:
            plot_bar(range(len(eig_imag)), eig_imag, color='#0081bf', label="imag", alpha=0.7, title=title, xlabel=xlabel, ylabel=ylabel)


def save_func(func, filename, domain, args=(), func_err=None, args_err=(), grad = None, header=None, 
              npoints=1000, **kwargs):

    checkDomain(domain)
    xmin = domain[0] 
    xmax = domain[1] 

    if xmin is None:
        for line in plt.gca().lines:
            xmin_new = np.min(line.get_xdata())
            if xmin is None:
                xmin = xmin_new
            if xmin_new < xmin:
                xmin = xmin_new
    if xmax is None:
        for line in plt.gca().lines:
            xmax_new = np.max(line.get_xdata())
            if xmax is None:
                xmax = xmax_new
            if xmax_new > xmax:
                xmax = xmax_new

    xdata = np.arange(xmin, xmax, (xmax - xmin) / npoints)
    ydata = func(xdata, args)

    createFilePath(filename)

    if func_err is not None:
        ydata_err = func_err(xdata, args_err)
        writeTable(filename,xdata,ydata,ydata_err,header=header,**kwargs)

    elif len(args_err) > 0:
        if (grad is None):
            logger.warn("Used numerical derivative!")

        # Arguments that are part of the error propagation
        tmp_args = tuple(args)[0:len(args_err)]

        # Optional arguments that are constant and, therefore, not part of the error propagation
        tmp_opt = tuple(args)[len(args_err):]

        ydata_err = error_prop_func(xdata, func, tmp_args, args_err, grad=grad, args=tmp_opt)

        writeTable(filename,xdata,ydata,ydata_err,header=header,**kwargs)

    else:
        writeTable(filename,xdata,ydata,header=header,**kwargs)


def do_fit(func, xdata, ydata, edata=None, start_params=None, priorval=None, priorsigma=None,
           algorithm="curve_fit", detailedInfo=False, show_results=False, **kwargs):
    """ 
    Wrapper to fitter initialization and the fit in one step. See above for arguments. 
    """
    fit = Fitter(func, xdata, ydata, edata, **kwargs)
    return fit.do_fit(start_params=start_params, priorval=priorval, priorsigma=priorsigma, algorithm=algorithm,
                      detailedInfo=detailedInfo, show_results=show_results)


def try_fit(func, xdata, ydata, edata=None, start_params=None, priorval=None, priorsigma=None,
            algorithms=std_algs, detailedInfo=False, show_results=False, **kwargs):
    """ 
    Wrapper to fitter initialization and the fit in one step. See above for arguments. For historical reasons
    algorithms has no default values here. 
    """
    fit = Fitter(func, xdata, ydata, edata, **kwargs)
    return fit.try_fit(start_params=start_params, priorval=priorval, priorsigma=priorsigma, algorithms=algorithms,
                       detailedInfo=detailedInfo, show_results=show_results)



# 
# runningCoupling.py                                                               
# 
# D. Clarke
# 
# Python implementation of functions related to running coupling. Aims toward QCD, but implemented for
# SU(Nc) with arbitrary Nc and arbitrary Nf. The coefficients are given in terms of g(mu),
#
#     beta = -g^2 ( b0 g + b1 g^3 + ... ) 
#


import numpy as np
from scipy.special import zeta
from latqcdtools.base.check import checkType


ZETA_3 = zeta(3)


def CF(Nc) -> float:
    """ 
    Casimir operator of defining representation of SU(Nc).

    Args:
        Nc (int): Number of colors.

    Returns:
        float : CF 
    """
    checkType('int',Nc=Nc)
    return (Nc**2-1)/(2*Nc)


def CA(Nc) -> float:
    """ 
    Casimir operator of adjoint representation of SU(Nc).

    Args:
        Nc (int): Number of colors.

    Returns:
        float : CA 
    """
    checkType('int',Nc=Nc)
    return Nc 


# 10.1103/PhysRevLett.30.1346, 10.1103/PhysRevLett.30.1343
def b0(Nf,Nc=3) -> float:
    """ 
    Universal one-loop beta coefficient for SU(Nc), expansion in g.  

    Args:
        Nf (int): Number of active quark flavors. 
        Nc (int): Number of colors.

    Returns:
        float : b0 
    """
    checkType('int',Nf=Nf)
    return ( 11*CA(Nc)/3 - 2*Nf/3 )


def b1(Nf,Nc=3) -> float:
    """ 
    Universal two-loop beta coefficient for SU(Nc), expansion in g. 

    Args:
        Nf (int): Number of active quark flavors.
        Nc (int): Number of colors.

    Returns:
        float : b1 
    """
    checkType('int',Nf=Nf)
    return ( 34*CA(Nc)**2/3 - 2*CF(Nc)*Nf - 10*CA(Nc)*Nf/3 )


# Larin and Vermaseren, Phys Lett B 303 (1993) 334-336
def b2_dimreg_MSbar(Nf,Nc=3) -> float:
    """ 
    Three-loop beta coefficient for SU(Nc) using dimensional regularization in MS-bar scheme. 

    Args:
        Nf (int): Number of active quark flavors.
        Nc (int): Number of colors.

    Returns:
        float : b2_MS-bar (dim reg) 
    """
    checkType('int',Nf=Nf)
    return ( 2857*CA(Nc)**3/54 + Nf   *(CF(Nc)**2-205*CF(Nc)*CA(Nc)/18-1415*CA(Nc)**2/54)
                               + Nf**2*(11*CF(Nc)/9+79*CA(Nc)/54)
            )


# van Ritbergen, Larin, and Vermaseren, Phys Lett B 400 (1997) 379-384
def b3_dimreg_MSbar(Nf,Nc=3) -> float:
    """ 
    Four-loop beta coefficient for SU(Nc) using dimensional regularization in MS-bar scheme. 

    Args:
        Nf (int): Number of active quark flavors.
        Nc (int): Number of colors.

    Returns:
        float : b3_MS-bar (dim reg) 
    """
    checkType('int',Nf=Nf)
    return ( CA(Nc)**4*(150653/486-44*ZETA_3/9) + Nc**2*(Nc**2+36)/24*(704*ZETA_3/3-80/9)
                + Nf   *(CA(Nc)**3/2*(136*ZETA_3/3-39143/81) 
                          + CA(Nc)**2*CF(Nc)/2*(7073/243-656*ZETA_3/9) 
                          + CA(Nc)*CF(Nc)**2/2*(352*ZETA_3/9-4204/27) 
                          + 46*CF(Nc)**3/2 
                          + Nc*(Nc**2+6)/48*(512/9-1664*ZETA_3/3)
                         )
                + Nf**2*(CA(Nc)**2/4*(224*ZETA_3/9+7930/81)
                          + CF(Nc)**2/4*(1352/27-704*ZETA_3/9)
                          + CA(Nc)*CF(Nc)/4*(17152/243+448*ZETA_3/9)
                          + (Nc**4-6*Nc**2+18)/(96*Nc**2)*(512*ZETA_3/3-704/9)
                         )
                + Nf**3*(424*CA(Nc)/(8*243)+1232*CF(Nc)/(8*243)
                         )
            )


def beta_func(beta,Nf=3) -> float:
    """ 
    QCD asymptotic scaling relation to two loops.

    Args:
        beta (float-like)
        Nf (int, optional): Number of active quark flavors. Defaults to 3.

    Returns:
        float: f_as(beta) 
    """
    checkType('int',Nf=Nf)
    B0 = b0(Nf)/(4*np.pi)**2
    B1 = b1(Nf)/(4*np.pi)**4
    return (B0*10/beta)**(-B1/(2*B0**2)) * np.exp(-beta/(20*B0))



from setuptools import setup, find_packages


def read_requirements():
    """Read requirements.txt and return a list of dependencies."""
    with open("/home/dclarke/GitHub/AnalysisToolbox/requirements.txt", "r") as fh:
        return fh.read().splitlines()


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="latqcdtools",
    version="1.3.1",
    author="D. A. Clarke",
    author_email="clarke.davida@gmail.com",
    description="A set of Python tools for statistically analyzing correlated data. This includes aspects of lattice QCD applications related to QCD phenomenology.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LatticeQCD/AnalysisToolbox",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=read_requirements(),
)



#
# ideal.py 
#
# D. Clarke 
#
# Methods for ideal gas calculations. 
#


import numpy as np
import sympy
from latqcdtools.base.check import checkType
import latqcdtools.base.logger as logger

class idealGas:
    """ 
    Class for ideal, massless gas of quarks, zeroth order in the coupling. Fermion chemical
    potentials are added in the order u-->d-->s-->c.

    Args:
        Nc (int): Number of colors. 
        Nf (int): Number of fermion flavors. 
    """

    def __init__(self,Nc,Nf): 
        checkType('int',Nc=Nc)
        checkType('int',Nf=Nf)
        if Nf >= 4:
            logger.TBRaise('Only B, Q, S, C explicitly coded in.')
        if Nc < 1:
            logger.TBRaise('Must have positive number of colors.')
        self.T, self.muB, self.muS, self.muQ, self.muC = sympy.symbols('T muB muS muQ muC') 
        self.Nf   = Nf
        self.Nc   = Nc
        self.muu  = (self.muB + 2*self.muQ)/3
        self.mud  = (self.muB -   self.muQ)/3
        self.mus  = (self.muB -   self.muQ)/3 - self.muS
        self.muc  = (self.muB + 2*self.muQ)/3 + self.muC
        # eq (8.42) and (8.43) in Kapusta and Gale's "Finite temperature field theory: 
        # Principles and Applications" (2006)
        self.Psym = (np.pi**2/45)*(self.Nc**2-1)*self.T**4 
        if Nf > 0:  
            self.Psym += self.Nc*( 7*np.pi**2*self.T**4/180 + self.muu**2*self.T**2/6 + self.muu**4/(12*np.pi**2) )
        if Nf > 1:  
            self.Psym += self.Nc*( 7*np.pi**2*self.T**4/180 + self.mud**2*self.T**2/6 + self.mud**4/(12*np.pi**2) )
        if Nf > 2:  
            self.Psym += self.Nc*( 7*np.pi**2*self.T**4/180 + self.mus**2*self.T**2/6 + self.mus**4/(12*np.pi**2) )
        if Nf > 3:  
            self.Psym += self.Nc*( 7*np.pi**2*self.T**4/180 + self.muc**2*self.T**2/6 + self.muc**4/(12*np.pi**2) )

    def __repr__(self) -> str:
        return "idealGas"

    def P(self, T, muB=0., muS=0., muQ=0., muC=0.):
        """ 
        Unitful pressure. 
        """
        values = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
        return np.float128(self.Psym.subs(values).evalf())

    def gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB=0., muQ=0., muS=0., muC=0.):
        """
        Unitful cumulant.
        """
        chi = sympy.diff(self.Psym, self.muB, B_order)
        chi = sympy.diff(chi      , self.muQ, Q_order)
        chi = sympy.diff(chi      , self.muS, S_order)
        chi = sympy.diff(chi      , self.muC, C_order)
        values = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
        return np.float128(chi.subs(values).evalf())

    def S(self, T, muB=0., muS=0., muQ=0., muC=0.):
        """ 
        Unitful entropy. 
        """
        entropy = sympy.diff(self.Psym, self.T, 1)
        values = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
        return np.float128(entropy.subs(values).evalf())

#    def E(self, T, muB=0., muS=0., muQ=0., muC=0.):
#        """
#        Unitful energy density.
#        """
#        epsilon = self.T**2*sympy.diff(self.Psym/self.T, self.T, 1)
#        values  = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
#        return float(epsilon.subs(values).evalf())

    def ddT_gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB=0., muQ=0., muS=0., muC=0.):
        chi     = sympy.diff(self.Psym, self.muB, B_order)
        chi     = sympy.diff(chi      , self.muQ, Q_order)
        chi     = sympy.diff(chi      , self.muS, S_order)
        chi     = sympy.diff(chi      , self.muC, C_order)
        ddT_chi = sympy.diff(chi      , self.T  , 1      )
        values = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
        return np.float128(ddT_chi.subs(values).evalf())

    def d2dT2_gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB=0., muQ=0., muS=0., muC=0.):
        chi       = sympy.diff(self.Psym, self.muB, B_order)
        chi       = sympy.diff(chi      , self.muQ, Q_order)
        chi       = sympy.diff(chi      , self.muS, S_order)
        chi       = sympy.diff(chi      , self.muC, C_order)
        d2dT2_chi = sympy.diff(chi      , self.T  , 2      )
        values = {self.T: T, self.muB: muB, self.muQ: muQ, self.muS: muS, self.muC: muC} 
        return np.float128(d2dT2_chi.subs(values).evalf())




# 
# autocorrelation.py                                                               
#
# D. Clarke
# 
# Calculation of integrated autocorrelation time. The main methods have been adapted from software of
# Bernd Berg, Markov Chain Monte Carlo and their Statistical Analysis, World Scientific, 2004, ISBN=978-981-3106-37-6
# 

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.statistics.statistics import std_mean, std_err, checkTS
from latqcdtools.base.plotting import plt, clearPlot, plot_dots, set_params
from latqcdtools.statistics.jackknife import jackknife 
from latqcdtools.base.readWrite import writeTable
from latqcdtools.base.check import checkType


def remove1Jackknife(ts) -> np.ndarray:
    """ 
    Create remove-1 jackknife list from 1-d series.

    Args:
        ts (array-like): time series 

    Returns:
        np.array: 1-d array of jackknife means 
    """
    checkTS(ts)
    return jackknife(np.mean,ts,numb_blocks=len(ts),return_sample=True,conf_axis=0)[0]


def tauint(nt,ts,xhat = None) -> np.ndarray:
    """ 
    Given a time series, calculate estimators for its integrated autocorrelation time  at each Markov time separation.

    INPUT:
         nt--The largest you think tau_int could be.
         ts--Time series array of measurments. Must be taken from equilibrium ensemble so that
             time translation invariance holds. List must be in order of Markov chain generation.
       xhat--True mean of time series (if you know it).

    OUTPUT:
      acint--List of integrated autocorrelation times. 
      """
    checkTS(ts)
    ndat=len(ts)
    if nt>=ndat:
        logger.TBRaise("Need nt<ndat.")
    if xhat is not None:
        x=xhat
    else:
        x=std_mean(ts)
    # Create array of autocovariance
    acov=[]
    for it in range(nt+1):
        numt=ndat-it # number of pairs of time series elements separated by computer time it
        c_it=0.
        for i in range(numt):
            c_it=c_it+(ts[i]-x)*(ts[i+it]-x)
        c_it=c_it/numt
        # Bias correction. This is needed for c(t) to be correct in the limit of uncorrelated data. In principle
        # you don't know ahead of time whether your raw data are effectively correlated. The factor is ndat rather than
        # ndat-it because ndat data points were used to calculate x, which is what matters for the bias.
        if xhat is None:
            c_it=c_it*ndat/(ndat-1.)
        acov.append(c_it)
    acov=np.array(acov)
    # Calculate integrated autocorrelation time. This is equivalent to the following code:
    #    acint=[1.]
    #    for it in range(1,nt+1):
    #        acint.append( acint[it-1] + 2.*acov[it]/acov[0] )
    acint     = np.zeros(nt + 1)
    acint[0]  = 1
    acint[1:] = 1 + 2*np.cumsum(acov[1:]/acov[0])
    return acint


def tauintj(nt,nbins,ts,xhat = None) -> np.ndarray:
    """ 
    Given a time series, calculate jackknife bins of integrated autocorrelation time for each Markov time separation.

    INPUT:
          nt--The largest nt at which you think your estimate for tau_int could lie.
       nbins--The number of jackknife bins.
          ts--Time series array of measurements. Must be taken from equilibrium ensemble so that
              time translation invariance holds. List must be in order of markov chain generation
        xhat--True mean of time series (if you know it).

    OUTPUT:
      acintj--2D list indexed by time, then bin number acintj[it][ibin] 
      """
    checkType('int',nt=nt)
    checkType('int',nbins=nbins)
    checkTS(ts)
    ndat=len(ts)
    if not 1<nbins<ndat:
        logger.TBRaise("Need 1<nbins<ndat. nbins, ndat =",nbins,ndat)
    if xhat is not None:
        x=xhat
    else:
        x=std_mean(ts)
    # For each it, create a list of nbins jackknife measurements of the autocovariance. The measurements are an average
    # over binsize data with separation it. Everything is stored in a 2D array acorj[it][ibin].
    acorj=[]
    for it in range(nt+1):
        numt=ndat-it
        binsize=int(numt/nbins)
        acor=np.zeros(nbins)
        # This is a check that we can't spill into the last bin
        itmax=nbins*binsize-binsize
        if it>=itmax:
            logger.TBRaise("it>=itmax.")
        for ibin in range(nbins):
            i1=ibin*binsize
            i2=(ibin+1)*binsize-1
            acor[ibin] = np.sum( (ts[i1:i2+1] - x)*(ts[i1+it:i2+1+it] - x) )/binsize
        # Bias correction
        if xhat is None:
            acor=acor*ndat/(ndat-1.)
        acorj.append(remove1Jackknife(acor))
    acorj = np.array(acorj)
    # The following code is equivlent to this (but faster since it uses numpy)
    #    acintj=[[1.]*nbins]
    #    for it in range(1,nt+1):
    #        tauintbins=[]
    #        for ibin in range(nbins):
    #            tauintbins.append(acintj[it-1][ibin]+2.*acorj[it][ibin]/acorj[0][ibin])
    #        acintj.append(tauintbins)
    acintj = np.ones((nt+1,nbins))
    increments = 2.*acorj[1:]/acorj[0]
    acintj[1:] = np.cumsum(increments, axis=0) + 1
    return acintj


def getTauInt(ts, nbins, tpickMax, acoutfileName = None, showPlot = False):
    """ 
    Given a time series, return estimates for the integrated autocorrelation time and its error.

    INPUT:
         tpickMax--The largest nt where you think your estimate might become unreliable.
            nbins--The number of jackknife bins (for estimating the error in tau_int)
               ts--Time series array of measurements. Must be taken from equilibrium ensemble so that
                   time translation invariance holds. List must be in order of markov chain generation

    OUTPUT:
          tau_int--Estimate for integrated autocorrelation time.
         tau_inte--Its (jackknife) error bar.
           itpick--The Monte Carlo separation at which this method found its estimate for tau_int. 
    """
    checkTS(ts)

    # Get integrated autocorrelation time list and corresponding jackknife list.
    acint  = tauint (tpickMax,ts,)
    acintj = tauintj(tpickMax,nbins,ts)

    # This block outputs time, tau_int and tau_int error, then gives an estimate of tau_int. When tau_int
    # decreases for the first time, we call that our estimate. This is because we know tau_int should be a 
    # monotonically increasing function of t.
    lmonoton=True
    tau_int=0.
    tau_inte=-1
    itpick=-1
    its = np.arange(tpickMax+1)
    ace = std_err(acintj,axis=1)
    for it in its:
        if lmonoton:
            if not acint[it]<tau_int:
                tau_int=acint[it]
                tau_inte=ace[it]
                itpick=it
            else:  # acint[it] < tau_int ==> tau_int decreased
                lmonoton=False

    if acoutfileName is not None:
        writeTable(acoutfileName,its,acint,ace,header=['t','tau(t)','tau_err(t)'])

    if showPlot:
        clearPlot()
        plot_dots(its,acint,ace,color='black',marker=None)
        set_params(xlabel='conf', ylabel='$\\tau_{\\rm int}$')
        plt.show()

    return tau_int, tau_inte, itpick



# 
# simulationManagement.py                                                               
# 
# D. Clarke
# 
# This is some Python code to help assist with managing lattice simulations. 
# 

from latqcdtools.base.check import checkType
import latqcdtools.base.logger as logger
from latqcdtools.base.plotting import plt, plot_lines, plot_hist, set_params, clearPlot,\
    saveFigure
from latqcdtools.statistics.statistics import std_mean, std_err, checkTS, KSTest_1side
import scipy as sp
from latqcdtools.base.utilities import ls, naturalSort
from latqcdtools.math.math import rel_check

def countConfigurations(targetFolder,name,delimiter='.'):
    """
    Count the number of configurations in the target folder.

    Args:
        targetFolder (str)
        name (str): Assume configuration name has form name.###. 
        delimiter (str, optional): The delimiter between name and ###. Defaults to '.'.

    Returns:
        int: number of configurations in targetFolder 
    """
    checkType(str,targetFolder=targetFolder)
    checkType(str,name=name)
    checkType(str,delimiter=delimiter)
    confs=[]
    files=ls(f'{targetFolder}/{name}*')
    for f in files:
        try:
            _, confno = f.rsplit(delimiter,1)
            confs.append(int(confno))
        except ValueError as e:
            logger.TBRaise(f'Configuration names in {targetFolder} must end in int. file={f}, delimiter="{delimiter}"',exception=e)
    logger.info(f'{targetFolder}:  min={min(confs)}  max={max(confs)}')
    return len(confs)


def analyzeChain(MCtime,measurements,obslabel=None,MClabel=None,KScutoff=0.05,
                 showPlots=False,savePlots=False,plotNamePrefix=None,**plotargs):
    """
    Do some basic analysis of a MCMC time series of measurements. We check whether the data
    are distributed normally. Optionally you can plot the time series and/or a histogram.

    Args:
        MCtime (array-like): Array indexing the MC time. 
        measurements (array-like): Time series of measurements.
        obslabel (str, optional): Label for observable in plots. Defaults to None.
        MClabel (str, optional): Label for x-axis of time series plot. Defaults to None.
        KScutoff (float, optional): Threshold below which "not normal enough". Defaults to 0.05.
        showPlots (bool, optional): Defaults to False.
        savePlots (bool, optional): Defaults to False.
        plotNamePrefix (str, optional): Prefix of plot names. Defaults to None.
    """

    checkTS(MCtime)
    checkTS(measurements)
    checkType('real',KScutoff=KScutoff)

    if len(plotargs)>0 and showPlots==False and savePlots==False:
        logger.TBRaise('Passed plotting keywords without wanting to plot.')
    if savePlots and (plotNamePrefix is None):
        logger.TBRaise('Please specify a name prefix for the TS and histogram plots.')

    if savePlots or showPlots:
        if MClabel is None:
            logger.TBRaise('Plotting without an observable label.')
        checkType(str,MClabel=MClabel)
        plot_lines(MCtime,measurements,marker=None,color='black')
        set_params(xlabel=MClabel,ylabel=obslabel,**plotargs)
    if savePlots:
        checkType(str,plotNamePrefix=plotNamePrefix)
        saveFigure(plotNamePrefix+'TS.pdf')
    if showPlots:
        plt.show()

    if savePlots or showPlots:
        clearPlot()
        plot_hist(measurements)
        set_params(xlabel=obslabel,**plotargs)
    if savePlots:
        saveFigure(plotNamePrefix+'hist.pdf')
    if showPlots:
        plt.show()
    clearPlot()

    mean = std_mean(measurements)
    err  = std_err(measurements)
    normalCDF = sp.stats.norm(loc=mean,scale=err).cdf

    if KSTest_1side(measurements,normalCDF)<KScutoff:
        logger.warn('Measurement distribution not consistent with Gaussian')



_allowedEnsKinds = ['conf','meas','conf/meas','seed','seed/meas']
_allowedEnsKeys  = ['kind',  # configurations? measurements?
                    'date',  # last date you checked it
                    'Nt',
                    'Ns', 
                    'beta',
                    'Nf', 
                    'ml',    # Nf=2+1: light mass 
                    'ms',    # Nf=2+1: strange mass
                    'mc',    # Nf=2+1: charm mass
                    'mf',    # Nf=N: fermion mass
                    'mpre',  # Nf=N: preconditioner mass
                    'location', 
                    'folder', 
                    'notes', 
                    'creator'
                    ]
_allowedEnsNf    = ['2+1','3','5','2+1+1']


class ensemble:

    """ 
    An ensemble object. This is to help you organize where data are saved
    """

    def __repr__(self) -> str:
        return "ensemble"

    def __init__(self,**kwargs):
        self._ens = {
            'kind'    :None,
            'date'    :None, 
            'Nt'      :None, 
            'Ns'      :None, 
            'beta'    :None, 
            'Nf'      :None,
            'ml'      :None, 
            'ms'      :None, 
            'mc'      :None, 
            'mf'      :None, 
            'mpre'    :None,
            'ml/ms'   :None, 
            'location':None, 
            'folder'  :None, 
            'notes'   :None,
            'creator' :None
        }
        for key in kwargs:
            if key not in _allowedEnsKeys:
                logger.warn(f'Unknown ensemble keyword {key}')
            else:
                self._ens[key] = kwargs[key] 
        lcomplete = True
        if self._ens['kind'] is None:
            lcomplete = False
        if self._ens['Nt'] is None:
            lcomplete = False
        if self._ens['Nf'] is None:
            lcomplete = False
        if self._ens['beta'] is None:
            lcomplete = False
        if not lcomplete:
            logger.TBRaise('Must minimally specify Nt, Nf, beta, and kind.')
        if self._ens['kind'] not in _allowedEnsKinds: 
            logger.TBRaise('Unrecognized kind',self._ens['kind'])
        if self._ens['Nf'] not in _allowedEnsNf: 
            logger.TBRaise('Unrecognized Nf',self._ens['Nf'])
        if self._ens['Nf']=='2+1' or self._ens['Nf']=='2+1+1':
            if (self._ens['mf'] is not None) or (self._ens['mpre'] is not None):
                logger.TBRaise('Detected degenerate Nf param with Nf=2+1(+1)')
        if self._ens['Nf']=='3' or self._ens['Nf']=='5':
            if (self._ens['ms'] is not None) or (self._ens['ms'] is not None) or (self._ens['mc'] is not None):
                logger.TBRaise('Detected Nf=2+1(+1) param with degenerate Nf')
        if self._ens['mc'] is not None:
            if self._ens['Nf']!='2+1+1':
                logger.TBRaise('Specified charm mass without Nf=2+1+1')
        if (self._ens['ms'] is not None) and (self._ens['ml'] is not None):
            msml = self._ens['ms']/self._ens['ml']
            if rel_check(msml,27,0.01):
                self._ens['ml/ms'] = 'phys'
            if rel_check(msml,80,0.01):
                self._ens['ml/ms'] = '1/80'

    def get(self,key):
        if self._ens[key] is not None:
            return self._ens[key]
        else:
            return '-'

    def accessor(self) -> dict:
        return self._ens


class repository():

    """
    This reposistory holds ensemble objects.
    """

    def __init__(self):
        self._repo = []

    def accessor(self) -> list:
        return self._repo

    def append(self,ens):
        checkType(ensemble,ens=ens) 
        self._repo.append(ens)

    def search(self,key,val,verbose=True):
        matches = repository()
        nmatches = 0
        for ens in self._repo: 
            if ens.get(key)==val:
                matches.append(ens)
                nmatches += 1
        if verbose:
            logger.info(f'Found {nmatches} matches:')
            matches.list(verbose=True)
        return matches

    def list(self,verbose=False):
        if verbose:
            logger.info()
            for ens in self._repo:
                for key in _allowedEnsKeys: 
                    if ens.get(key) is not None:
                        logger.info(f'  {key} = {ens.get(key)}')
                logger.info()
        logger.info(f'N_ens = {len(self._repo)}')

    def findUniqueKeys(self,key) -> set:
        ret = []
        for ens in self._repo:
            ret.append(ens.get(key))
        return set(ret)

    def table(self):
        logger.info(f'{"Nf":<6} {"ml/ms":<6} {"Nt":<4} {"beta":<8} {"Ns":<4} {"ml (mf)":<12} {"ms (mpre)":<10} {"mc":<8} {"kind":<12} {"location":<10} folder')
        logger.info()

        # This loop controls the order in which entries appear in the table.
        Nfs = self.findUniqueKeys('Nf') 
        for Nf in naturalSort(Nfs):
            repoNf = self.search('Nf',Nf,verbose=False)
            mqs = repoNf.findUniqueKeys('ml/ms')
            for mq in mqs:
                repomq = repoNf.search('ml/ms',mq,verbose=False)
                Nts = repomq.findUniqueKeys('Nt') 
                for Nt in sorted(Nts):
                    repoNt = repomq.search('Nt',Nt,verbose=False)
                    betas = repoNt.findUniqueKeys('beta')
                    for beta in sorted(betas):
                        repoBeta = repoNt.search('beta',beta,verbose=False)
                        Nss = repoBeta.findUniqueKeys('Ns')
                        for Ns in sorted(Nss):
                            repoNs = repoBeta.search('Ns',Ns,verbose=False)
                            for ens in repoNs.accessor():
                                if Nf=='2+1':
                                    logger.info(f"{Nf:<6} {ens.get('ml/ms'):<6} {Nt:<4} {beta:<8} {Ns:<4} {ens.get('ml'):<12} {ens.get('ms'):<10} {'-':<8} {ens.get('kind'):<12} {ens.get('location'):<10} {ens.get('folder')}")
                                elif Nf=='2+1+1':
                                    logger.info(f"{Nf:<6} {ens.get('ml/ms'):<6} {Nt:<4} {beta:<8} {Ns:<4} {ens.get('ml'):<12} {ens.get('ms'):<10} {ens.get('mc'):<8} {ens.get('kind'):<12} {ens.get('location'):<10} {ens.get('folder')}")
                                elif Nf=='3' or Nf=='5':
                                    logger.info(f"{Nf:<6} {ens.get('ml/ms'):<6} {Nt:<4} {beta:<8} {Ns:<4} {ens.get('mf'):<12} {ens.get('mpre'):<10} {'-':<8} {ens.get('kind'):<12} {ens.get('location'):<10} {ens.get('folder')}")


#
# check.py
#
# D. Clarke
#
# Warning and error control, along with methods to check internal code consistency. 
#

import warnings, inspect
import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.utilities import envector, isArrayLike, isIntType, isScalar


# This warning is on by default, which worries that we may lose precision when using complex numbers. But
# I think in the instances I use complex numbers, they cannot really be avoided.
try:
    warnings.filterwarnings("ignore", category=np.ComplexWarning)
except AttributeError:
    pass


# I want solvers to try other strategies when they hit RuntimeWarnings.
warnings.filterwarnings("error", category=RuntimeWarning)


class DivideByZeroError(Exception): pass
class UnderflowError(Exception): pass
class InvalidValueError(Exception): pass


CATCHUNDERFLOW    = True 
CATCHOVERFLOW     = True 
CATCHDIVIDEBYZERO = True
CATCHINVALIDVALUE = True


def err_handler(err, flag):
    """ 
    This method lets us control in detail how different types of errors are treated. 
    """
    global CATCHUNDERFLOW
    global CATCHOVERFLOW
    global CATCHDIVIDEBYZERO
    global CATCHINVALIDVALUE
    if flag == 1:
        if CATCHDIVIDEBYZERO:
            raise DivideByZeroError(err)
        else:
            pass
    elif flag == 2:
        if CATCHOVERFLOW:
            raise OverflowError
        else:
            pass
    elif flag == 4:
        if CATCHUNDERFLOW: 
            raise UnderflowError(err)
        else:
            pass
    elif flag == 6:
        if CATCHOVERFLOW: 
            raise OverflowError(err)
        else:
            pass
    elif flag == 8:
        if CATCHINVALIDVALUE:
            raise InvalidValueError(err)
        else:
            pass
    elif flag == 9:
        if CATCHDIVIDEBYZERO:
            raise DivideByZeroError(err)
        else:
            pass
    elif flag == 10:
        if CATCHOVERFLOW: 
            raise OverflowError(err)
        else:
            pass
    elif flag == 12:
        if CATCHUNDERFLOW: 
            raise UnderflowError(err)
        else:
            pass
    else:
        logger.TBRaise('Encountered unknown exception',err,'with flag',flag)


np.seterrcall(err_handler)
np.seterr(all='call')


def ignoreUnderflow():
    """ 
    Turn off underflow crashes. 
    """
    global CATCHUNDERFLOW
    CATCHUNDERFLOW = False
    logger.warn("Underflow behavior set to pass.")


def ignoreOverflow():
    """ 
    Turn off overflow crashes. 
    """
    global CATCHOVERFLOW
    CATCHOVERFLOW = False
    logger.warn("Overflow behavior set to pass.")


def ignoreDivideByZero():
    """ 
    Turn off zero division crashes. 
    """
    global CATCHDIVIDEBYZERO
    CATCHDIVIDEBYZERO = False
    logger.warn("Zero division behavior set to pass.")


def ignoreInvalidValue():
    """ 
    Turn off invalid value crashes. 
    """
    global CATCHINVALIDVALUE
    CATCHINVALIDVALUE = False
    logger.warn("Invalid value behavior set to pass.")


def checkType(expectedType,**kwargs):
    """ 
    Check the type of an object. If it thinks the type is wrong, it will tell you what the
    name of obj is (as you named it in your code) along with its type and what was expected.
    Grabbing the name doesn't work if you pass him a dictionary element like myDict['key'];
    it can only tell the name myDict. One could use type hints, but at the time of writing,
    type hints will not necessarily crash the program, which I want.

    Args:
        obj (obj)
        expectedType (type): what type do you expect? Also accepts "array", "real", "int", and "scalar".
    """
    if len(kwargs)!=1:
        logger.TBRaise('Call like checkType(expectedtype, var=value)') 
    objName = list(kwargs.keys())[0]
    if not isinstance(objName,str):
        logger.TBRaise('Call like checkType(expectedtype, var=value)') 
    obj = kwargs[objName]
    if expectedType=="array":
        if not isArrayLike(obj):
            if type(obj)==list or type(obj)==np.ndarray or type(obj)==tuple:
                logger.TBRaise('Received empty',type(obj),'for',objName,frame=3)
            else:
                logger.TBRaise('Expected array-like object for',objName,'but received',type(obj),frame=3)
    elif expectedType=="scalar":
        if not isScalar(obj): 
            logger.TBRaise('Expected scalar object for',objName,'but received',type(obj),frame=3)
    elif expectedType=="real":
        if not isScalar(obj): 
            logger.TBRaise('Expected real scalar object for',objName,'but received',type(obj),frame=3)
        elif obj.imag>0:
            logger.TBRaise('Expected real scalar object',objName,'has nonzero imaginary part',frame=3)
    elif expectedType=="int":
        if not isIntType(obj):
            logger.TBRaise('Expected int object for',objName,'but received',type(obj),frame=3)
    else:
        if not isinstance(obj,expectedType):
            logger.TBRaise('Expected type',expectedType,'for',objName,'but received',type(obj),frame=3)


def checkDomain(obj, expectedDomain):
    """ 
    Check that obj lies in expectedDomain.

    Args:
        obj (obj)
        expectedDomain (array-like): collection of values obj is allowed to take
    """
    checkType("array",expectedDomain=expectedDomain)
    calling_frame = inspect.currentframe().f_back
    locals_dict = calling_frame.f_locals
    for var_name, _ in locals_dict.items():
        objName = var_name  
    if not obj in expectedDomain:
        logger.TBRaise('Expected',objName,'to be one of',expectedDomain,frame=3)


def checkEqualLengths(*args):
    """ 
    Check that all array-like objects passed have the same length. 
    """
    length = len(envector(args[0]))
    for i in range(len(args)):
        if args[i] is not None:
            len_i = len(envector(args[i]))
            if len_i != length:
                logger.info(length, len(envector(args[i])))
                logger.TBRaise(f'Array length mismatch detected on array {i}. len, len[i] = {length}, {len_i}',frame=3)


def checkExtension(filename,extension):
    """ 
    Check the extension of a file

    Args:
        filename (str)
        extension (str)
    """
    if not filename.endswith(extension):
        logger.TBRaise('Expected a',extension,'file.')



# 
# cleanData.py
# 
# D. Clarke
# 
# Some methods for cleaning up numpy arrays. You might use this e.g. to restrict data to certain intervals, or to
# exclude NaN and Inf. These are essentially wrappers for array masks, which I find hard to read.
#


import numpy as np
from latqcdtools.base.check import checkType
import latqcdtools.base.logger as logger


def deleteRow(array, row) -> np.ndarray:
    """
    Remove a row of a 2d np.ndarray.

    Args:
        array (np.ndarray)
        row (int): Remove this row 

    Returns:
        np.ndarray: array with row removed
    """
    checkType(np.ndarray,array=array)
    checkType('int',row=row)
    if array.ndim != 2:
        logger.TBRaise('Expected 2-d numpy array.')       
    return np.delete(array,row,0)


def deleteCol(array, col) -> np.ndarray:
    """
    Remove a column of a 2d np.ndarray.

    Args:
        array (np.ndarray)
        col (int): Remove this column 

    Returns:
        np.ndarray: array with row removed
    """
    checkType(np.ndarray,array=array)
    checkType('int',col=col)
    if array.ndim != 2:
        logger.TBRaise('Expected 2-d numpy array.')       
    return np.delete(array,col,1)


def clipRange(array, col = None, minVal=-np.inf, maxVal=np.inf) -> np.ndarray:
    """ 
    Throw out any elements of array that lie outside the interval (minVal,maxVal). Note this
    renders arrays finite. 
    """
    checkType(np.ndarray,array=array)
    if col is None:
        mask = np.logical_and( array[:]>minVal, array[:]<maxVal )
        return array[mask]
    else:
        if array.ndim<2:
            logger.TBRaise('Set col=None for 1-d numpy arrays.')
        mask = np.logical_and( array[col,:]>minVal, array[col,:]<maxVal )
        return array[:,mask]


def intersectAtCol(table1, table2, col):
    """ 
    Return only those rows of table1 and table2 that have identical elements in column col. 
    """
    checkType(np.ndarray, table1=table1)
    checkType(np.ndarray, table2=table2)
    checkType('int', col=col)
    mask1using2 = np.isin( table1[col,:], table2[col,:] )
    mask2using1 = np.isin( table2[col,:], table1[col,:] )
    return table1[:,mask1using2], table2[:,mask2using1]


def spliceAtCol(table1, table2, col, atVal) -> np.ndarray:
    """ 
    Assuming two tables table1 and table2 have common values in column col, create a new
    table, where table1 has corresponding entries less than atVal in col, and table 2
    has corresponding entries greater than atVal. 
    """
    checkType(np.ndarray,table1=table1)
    checkType(np.ndarray,table2=table2)
    checkType('int',col=col)
    mask1 = table1[col]<=atVal
    mask2 = table2[col]>atVal
    return np.column_stack( (table1[:,mask1], table2[:,mask2]) )


def restrictAtCol(table, col, atVal, rtol=None, atol=None) -> np.ndarray:
    """ 
    Return only those rows of table where col has exactly the value atVal. 
    """
    checkType(np.ndarray, table=table)
    checkType('int',col=col)
    checkType("scalar",atVal=atVal)
    if (rtol is None) and (atol is None):
        mask = np.equal(table[col,:],atVal)
    else:
        if rtol is None: # numpy defaults
            rtol=1e-5
        if atol is None:
            atol=1e-8
        mask = np.isclose(table[col,:],atVal,rtol=rtol,atol=atol)
    return table[:,mask]


def excludeAtCol(table, col=None, atVal=np.inf) -> np.ndarray:
    """ 
    Return everything except those rows of table where col has exactly the value atVal. 
    """
    checkType(np.ndarray,table=table)
    if col is None:
        mask = np.not_equal(table,atVal)
        return table[mask]
    else:
        checkType('int', col=col)
        mask = np.not_equal(table[col,:],atVal)
        return table[:,mask]



# 
# polyakovTools.py                                                               
# 
# D. Clarke
# 
# Some methods useful for calculating Polyakov loop observables. 
#

import numpy as np
from latqcdtools.base.check import checkType



class polyakovTools:

    """
    Methods for Polyakov loop observables. The methods all take as input measurements of the real and imaginary
    parts of the Polyakov loop as numpy arrays, polReIm.
    """

    def __init__(self, Nsigma, Ntau):
        checkType('int',Nsigma=Nsigma)
        checkType('int',Ntau=Ntau)
        self.Ns = Nsigma
        self.Nt = Ntau


    def __repr__(self) -> str:
        return "polyakovTools"


    # Absolute value of Polyakov loop

    def absPLoop(self, polReIm):
        """ 
        <|P|> 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        # Calculate |P|_ij on each configuration i belonging to the jackknife sample j, then
        # compute the mean |P|_j of that jackknife sample.
        return np.mean( np.sqrt(ReP**2+ImP**2) )


    def absPLoop2(self, polReIm):
        """ 
        <|P|>^2 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return np.mean( np.sqrt(ReP**2+ImP**2) )**2


    # Susceptibility and susceptibility-like observables

    def Suscept(self, polReIm):
        """ 
        chi_|P| 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return self.Ns**3*( np.mean(ReP**2+ImP**2)-np.mean(np.sqrt(ReP**2+ImP**2))**2 )


    def T3Suscept(self, polReIm):
        """ 
        T^3 chi_|P| 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return (self.Ns/self.Nt)**3*( np.mean(ReP**2+ImP**2)-np.mean(np.sqrt(ReP**2+ImP**2))**2 )


    def ReSuscept(self, polReIm):
        """ 
        chi_ReP 
        """
        ReP=polReIm[0]
        return self.Ns**3*( np.mean(ReP**2)-np.mean(ReP)**2 )


    def ImSuscept(self, polReIm):
        """ 
        chi_ImP 
        """
        ImP=polReIm[1]
        return self.Ns**3*( np.mean(ImP**2)-np.mean(ImP)**2 )


    def T3ReSuscept(self, polReIm):
        """ 
        T^3 chi_ReP 
        """
        ReP=polReIm[0]
        return (self.Ns/self.Nt)**3*( np.mean(ReP**2)-np.mean(ReP)**2 )


    def T3ImSuscept(self, polReIm):
        """ 
        T^3 chi_ImP 
        """
        ImP=polReIm[1]
        return (self.Ns/self.Nt)**3*( np.mean(ImP**2)-np.mean(ImP)**2 )


    def ReTimesIm(self, polReIm):
        """ 
        V <ReP ImP> 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return self.Ns**3*( np.mean(ReP*ImP) )


    def ReP2(self, polReIm):
        """ 
        V <ReP^2> 
        """
        ReP=polReIm[0]
        return self.Ns**3*( np.mean(ReP**2) )


    def ImP2(self, polReIm):
        """ 
        V <ImP^2> 
        """
        ImP=polReIm[1]
        return self.Ns**3*( np.mean(ImP**2) )


    def P2minReP2(self,polReIm):
        """ 
        V ( <|P|>^2 - <ReP>^2 ) 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return self.Ns**3*( np.mean(np.sqrt(ReP**2+ImP**2))**2-np.mean(ReP)**2 )


    # Susceptibility ratios


    def RatSuscFunction(self, polReIm):
        """ 
        R_T = chi_ImP / chi_ReP 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return (np.mean(ImP**2) - np.mean(ImP)**2) / (np.mean(ReP**2) - np.mean(ReP)**2)


    def RatSuscFunctionA(self, polReIm):
        """ 
        R_A = chi_|P| / chi_ReP 
        """
        ReP=polReIm[0]
        ImP=polReIm[1]
        return ( (np.mean(ReP**2+ImP**2)-np.mean(np.sqrt(ReP**2+ImP**2))**2) /
                 (np.mean(ReP**2)-np.mean(ReP)**2) )



#
# bootstr.py
#
# H. Sandmeyer, H.-T. Shu
#
# A parallelized bootstrap routine that can handle arbitrary return values of functions.
#


import numpy as np
from latqcdtools.statistics.statistics import meanArgWrapper, std_mean, std_dev, std_median, dev_by_dist
from latqcdtools.base.speedify import DEFAULTTHREADS, parallel_function_eval
from latqcdtools.base.initialize import DEFAULTSEED
from latqcdtools.base.check import checkType


def _autoSeed(seed) -> int:
    """ 
    We use seed=None to flag the seed should be automatically chosen. The problem is that we need
    seed to be an integer when enforcing that different bootstrap samples use different seeds. 
    """
    if seed is None:
        return np.random.randint(0,DEFAULTSEED)
    else:
        checkType('int',seed=seed)
        return seed


def recurs_append(data, sample_data, axis, conf_axis, sample_size, same_rand_for_obs, i, my_seed):
    """ 
    Recursive function to fill the sample. 
    """

    rng = np.random.default_rng(my_seed+i)
    if axis + 1 == conf_axis:
        numb_observe = len(data)
        if sample_size is None:
            sample_sizes = [ len(j) for j in data ]
        else:
            sample_sizes = [sample_size]*len(data)

        if not same_rand_for_obs:
            randints = [rng.integers(0, len(data[x]), size=sample_sizes[x]) for x in range(numb_observe)]
        else:
            tmp_rand = rng.integers(0, len(data[0]), size=sample_sizes[0])
            randints = [tmp_rand]*numb_observe
        for x in range(numb_observe):
            sample_data.append(np.array(data[x])[randints[x]])
        return

    else:
        for j in range(len(data)):
            sample_data.append([])
            recurs_append(data[j], sample_data[j], axis + 1, conf_axis, sample_size, same_rand_for_obs, i, my_seed)


class nimbleBoot:

    def __init__(self, func, data, numb_samples, sample_size, same_rand_for_obs, conf_axis, return_sample, seed,
                 err_by_dist, args, nproc):

        checkType('int',numb_samples=numb_samples)
        checkType(bool,same_rand_for_obs=same_rand_for_obs)
        checkType(bool,return_sample=return_sample)
        checkType(bool,err_by_dist=err_by_dist)
        checkType('int',nproc=nproc)
        self._func=func
        self._data=np.array(data)
        self._numb_samples=numb_samples
        if sample_size is not None:
            checkType('int',sample_size=sample_size)
        self._sample_size=sample_size
        self._same_rand_for_obs=same_rand_for_obs
        self._conf_axis=conf_axis
        self._return_sample=return_sample
        self._seed=_autoSeed(seed)
        self._err_by_dist=err_by_dist
        self._args=args
        self._nproc = nproc 

        if self._data.ndim == 1:
            self._conf_axis = 0

        self._sampleval = parallel_function_eval(self.getBootstrapEstimator,range(self._numb_samples),nproc=self._nproc,args=(self._seed,))

        if not self._err_by_dist:
            self._mean = std_mean(self._sampleval)
            self._error = std_dev(self._sampleval)
        else:
            self._mean = std_median(self._sampleval)
            self._error = dev_by_dist(self._sampleval)

    def __repr__(self) -> str:
        return "nimbleBoot"

    def getBootstrapEstimator(self,i,my_seed):
        sample_data = []
        if self._conf_axis == 0: # Case of one dimensional array is special
            rng = np.random.default_rng(my_seed+i)
            if self._sample_size is None:
                sample_size_tmp = len(self._data)
            else:
                sample_size_tmp = self._sample_size
            randints = rng.integers(0, len(self._data), size=sample_size_tmp)
            sample_data = self._data[randints]

        else:
            axis = 0
            recurs_append(self._data, sample_data, axis, self._conf_axis, self._sample_size, self._same_rand_for_obs, i, my_seed)

        sample_data = np.array(sample_data)

        return meanArgWrapper(self._func, sample_data, self._args)

    def getResults(self):
        if self._return_sample:
            return self._sampleval, self._mean, self._error
        else:
            return self._mean, self._error


def bootstr(func, data, numb_samples, sample_size = None, same_rand_for_obs = False, conf_axis = 1, return_sample = False,
            seed = None, err_by_dist = False, args=(), nproc=DEFAULTTHREADS):
    """
    Bootstrap for arbitrary functions. This routine resamples the data and passes them to the function in the same
    format as in the input. So the idea is to write a function that computes an observable from a given data set. This
    function can be put into this bootstrap routine and will get bootstrap samples as input. Based on the output of the
    function, the bootstrap mean and error are computed. The function may return multiple observables that are either
    scalars or numpy objects. You can pass a multidimensional object as data, but the bootstrap function has to know
    which axis should be resampled which is controlled by conf_axis (default = 0 for one dimensional arrays and
    default = 1 for higher order arrays.)

        Parameters
        ----------
        func : callable
            The function that calculates the observable

        data : array_like
            Input data

        numb_samples : integer
            Number of bootstrap samples

        sample_size : integer, optional, default = 0
            Size of sample

        same_rand_for_obs : boolean, optional, default = False
            Use the same random numbers for each observable accessed by index conf_axis - 1. Please note:
                - Objects that are accessed by an axis >= conf_axis + 1 do always have the same random numbers.
                - Objects that are accessed by axis conf_axis < conf_axis - 1 never share the same random numbers.

        conf_axis : integer, optional, default = 0 for dim(data) = 1 or default = 1 for dim(data) >= 2
            Axis that should be resampled

        return_sample : boolean, optional, default = False
            Along with the mean and the error also return the results from the individual samples

        seed: integer, optional, default = None
            seed for the random generator. If None, the default seed from numpy is used (probably from time)

        same_rand_for_obs : boolean, optional, default = False
            same random numbers per observable

        err_by_dist : boolean, optional, default = False
            Compute the error from the distribution using the median and the 68% quantile

        args : array_like or dict, default = ()
            optional arguments to be passed to func. If a dictionary the are passed as **args.

        nproc : integer
            Number of threads to use if you choose to parallelize. nproc=1 turns off parallelization.
    """
    bts = nimbleBoot(func, data, numb_samples, sample_size, same_rand_for_obs, conf_axis, return_sample, seed,
                     err_by_dist, args, nproc)
    return bts.getResults()


class nimbleGaussianBoot:

    def __init__(self, func, data, data_std_dev, numb_samples, sample_size, same_rand_for_obs, return_sample, seed,
                 err_by_dist, useCovariance, Covariance, args, nproc, asym_err):

        checkType('int',numb_samples=numb_samples)
        checkType('int',sample_size=sample_size)
        checkType(bool,same_rand_for_obs=same_rand_for_obs)
        checkType(bool,return_sample=return_sample)
        checkType(bool,err_by_dist=err_by_dist)
        checkType(bool,useCovariance=useCovariance)
        checkType('int',nproc=nproc)
        checkType(bool,asym_err=asym_err)
        self._func=func
        self._data=np.array(data)
        self._data_std_dev=np.array(data_std_dev)
        self._numb_samples=numb_samples
        self._sample_size=sample_size
        self._same_rand_for_obs=same_rand_for_obs
        self._return_sample=return_sample
        self._seed=_autoSeed(seed)
        checkType('int',seed=self._seed)
        self._err_by_dist=err_by_dist
        self._useCovariance=useCovariance
        self._Covariance=Covariance
        self._args=args
        self._numb_observe = len(data)
        self._nproc = nproc 

        self._sampleval=parallel_function_eval(self.getGaussianBootstrapEstimator,range(self._numb_samples),nproc=self._nproc,args=(self._seed,))

        if not self._err_by_dist:
            self._mean = std_mean(self._sampleval)
            self._error = std_dev(self._sampleval)
        else:
            self._mean = std_median(self._sampleval)
            self._error = dev_by_dist(self._sampleval, return_both_q=asym_err)

    def __repr__(self) -> str:
        return "nimbleGaussianBoot"

    def getGaussianBootstrapEstimator(self,i,my_seed):
        sample_data = []
        if not self._same_rand_for_obs:
            for k in range(self._numb_observe):
                rng=np.random.default_rng(my_seed+i+(k+1)*self._numb_samples)
                if not self._useCovariance:
                    if self._sample_size == 1:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k]))
                    else:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k], self._sample_size))
                else:
                    if self._sample_size == 1:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2)))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k]))
                    else:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2), self._sample_size))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k], self._sample_size))
        else:
            rng=np.random.default_rng(my_seed+i)
            for k in range(self._numb_observe):
                if not self._useCovariance:
                    if self._sample_size == 1:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k]))
                    else:
                        sample_data.append(rng.normal(self._data[k], self._data_std_dev[k], self._sample_size))
                else:
                    if self._sample_size == 1:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2)))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k]))
                    else:
                        if self._Covariance is None:
                            sample_data.append(rng.multivariate_normal(self._data[k], np.diag(self._data_std_dev[k]**2), self._sample_size))
                        else:
                            sample_data.append(rng.multivariate_normal(self._data[k], self._Covariance[k], self._sample_size))

        sample_data = np.array(sample_data)

        return meanArgWrapper(self._func,sample_data,self._args) 

    def getResults(self):
        if self._return_sample:
            return self._sampleval, self._mean, self._error
        else:
            return self._mean, self._error


def bootstr_from_gauss(func, data, data_std_dev, numb_samples, sample_size = 1, same_rand_for_obs = False,
                       return_sample = False, seed = None, err_by_dist = True, useCovariance = False,
                       Covariance = None, args = (), nproc = DEFAULTTHREADS, asym_err=False):
    """
    Same as standard bootstrap routine, but the data are generated by gaussian noise around the mean values in data.
    The width of the distribution is controlled by data_std_dev. Note, that the function has to average over samples.
    This means that data_std_dev should always be the standard deviation of a single measurement and not the standard
    deviation of a mean.
    """
    data = np.asarray(data)
    data_std_dev = np.asarray(data_std_dev)

    bts_gauss = nimbleGaussianBoot(func=func, data=data, data_std_dev=data_std_dev, numb_samples=numb_samples, 
                                   sample_size=sample_size, same_rand_for_obs=same_rand_for_obs,return_sample=return_sample, 
                                   seed=seed, err_by_dist=err_by_dist, useCovariance=useCovariance, Covariance=Covariance, 
                                   args=args, nproc=nproc, asym_err=asym_err)
    return bts_gauss.getResults()


def estimateCovariance(data_std_dev,numb_samples,seed=None):
    """ 
    Estimate a covariance matrix using data and their corresponding std_dev. The idea is to estimate this with a
    bootstrap under the assumption that each datum X_i is maximally correlated with every other datum X_j. Here
    data should be 1-dimensional.
    """
    checkType('int',numb_samples=numb_samples)
    rng = np.random.default_rng(seed=_autoSeed(seed))
    # The numpy outer and einsum below is equivalent to the following code:
    #    ndat = len(data_std_dev)
    #    cov = np.zeros((ndat,ndat))
    #    for ibs in range(numb_samples):
    #        scale = rng.normal(loc=0,scale=1)
    #        for i in range(ndat):
    #            for j in range(ndat):
    #                err_i     = scale*data_std_dev[i]
    #                err_j     = scale*data_std_dev[j]
    #                cov[i,j] += err_i*err_j
    scales = rng.normal(loc=0, scale=1, size=numb_samples)
    outer_product = np.outer(data_std_dev, data_std_dev)
    cov = np.einsum('k,ij->ij', scales**2, outer_product)
    return cov/numb_samples



# 
# lime.py                                                               
# 
# D. Clarke, X.-Y. Jin
# 
# Some tools for working with the LIME format. This is a port of X.-Y. Jin's original Python code,
# which can be found here: https://github.com/nftqcd/nthmc/blob/master/lib/fieldio.py. Information
# about the LIME format can be found here:
#
# To help with understanding LIME format, note the following organizational hierarchy:
#     message > record > word
# Message begin (MB) and message end (ME) flags have to be set according to the following rules:
#
# (M1)  MB=1 for the first record of the LIME file.
# (M2)  For any two consecutive records with ME flag me1 and MB flag mb2, respectively, the relation me1=mb2 must hold.
# (M3)  ME=1 for the last record of the LIME file.
#
# Special thanks to H. Simma for the explanation. Since we are not using the message organizational level, we
# will always set these flags to 1.
# 

import os, re, struct, zlib
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType

LIMEMAGIC = 0x456789ab


def limeHeader(mbeg, mend, size, type) -> bytes:
    """
    Get the header for a LIME record.

    Args:
        mbeg (bool): start flag 
        mend (bool): end flag 
        size (int): data length in bytes 
        type (bytes): description of data entry 

    Returns:
        bytes: LIME header byte sequence 
    """
    m = 0
    if mbeg:
        m = 1<<15 # starting flag in binary
    elif mend:
        m = 1<<14 # ending flag in binary

    # '> ensures big-endian byte order
    # i packs LIMEMAGIC as a 4-byte integer
    # h packs the integer 1 as a 2-byte short (stands in for version number)
    # H packs m as a 2-byte unsigned short
    # q packs size as an 8-byte long long
    # 128s packs type as a 128-byte string (padded with null bytes if necessary)
    return struct.pack('>ihHq128s', LIMEMAGIC, 1, m, size, type)


def scidacChecksum(latdata, vol, sitesize):
    """
    Compute SciDAC checksum for lattice data
    
    Args:
        latdata (bytes)
        vol (int): number of sites on lattice 
        sitesize (int): size in bytes of data at each site 

    Returns:
        checksum values (suma, sumb) 
    """
    suma = 0
    sumb = 0
    for site in range(vol):
        # Compute the CRC32 checksum for this site
        c = zlib.crc32(latdata[site*sitesize:(site+1)*sitesize]) & 0xffffffff
        s29 = site%29
        s31 = site%31
        # Updates sum by XORing it with c rotated left by s* bits and right by (32-s*) bits
        suma ^= (c<<s29 | c>>(32-s29)) & 0xffffffff
        sumb ^= (c<<s31 | c>>(32-s31)) & 0xffffffff
    return suma, sumb


def xmlFind(s,xmlTag) -> bytes:
    """
    Find substring in byte string s between xmlTags.
    
    Args:
        s (bytes)
        xmlTag (bytes)

    Returns:
        bytes: byte string between tags 
    """
    rs = '<'+xmlTag+r'>([^<]*)</'+xmlTag+'>'
    m = re.findall(rs.encode('utf-8'),s)
    if len(m)!=1:
        logger.TBRaise(f'finding record {xmlTag} in unsupported xml: {s}')
    return m[0]


def trimNull(byteString) -> bytes:
    """
    Remove null byte and everything after it from a byte string.

    Args:
        byteString (bytes)

    Returns:
        bytes: trimmed byte string 
    """
    n = byteString.find(b'\0')
    if n>=0:
        return byteString[:n]
    else:
        return byteString
    

def printLimeHeaders(filename):
    """
    Print the headers of a file in LIME format.

    Args:
        filename (str)
    """
    checkType(str,filename=filename)
    f = open(filename,'rb')  
    while True:
        header = f.read(144)
        if not header:
            break
        magi, vers, mbeg_end_res, size, type = struct.unpack('>ihHq128s',header)
        if magi!=LIMEMAGIC:
            logger.TBRaise(f'lime magic number does not match, got {magi}')
        logger.info('* HEADER')
        logger.info(f'   lime version: {vers}')
        mbeg = 1 & mbeg_end_res>>15
        mend = 1 & mbeg_end_res>>14
        mres = ~(3<<14) & mbeg_end_res
        logger.info(f'  mbeg mend res: {mbeg}, {mend}, {mres}')
        type = trimNull(type)
        logger.info(f'           size: {size} bytes')
        logger.info(f'           type: {type}')
        logger.info(f'       position: currently {f.tell()} bytes from start of file')
        if type==b'scidac-binary-data' or type==b'ildg-binary-data':
            next = (size+7)//8*8
        else:
            data = f.read(size)
            data = trimNull(data)
            logger.info(f'           data: {data}')
            next = 7-(size+7)%8
        f.seek(next,os.SEEK_CUR)
    f.close()
